# -*- coding: utf-8 -*-
"""
This module contain general info about lograptor package.
"""
#
# Copyright (C), 2011-2017, by SISSA - International School for Advanced Studies.
#
# This file is part of lograptor.
#
# Lograptor is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# file 'LICENSE' in the root directory of the present distribution
# for more details.
#
# @Author Davide Brunato <brunato@sissa.it>
#
__author__ = "Davide Brunato"
__copyright__ = "Copyright 2011-2017, SISSA"
__license__ = "LGPLv2.1+"
__version__ = "1.2.3"
__maintainer__ = "Davide Brunato"
__email__ = "brunato@sissa.it"
__status__ = "Production"
__description__ = ("Command-line utility for processing log files. "
                   "Produces matching outputs, data and reports.")
LONG_DESCRIPTION = """
lograptor is a tool which provides a command-line interface for system logs processing.

Pattern matching searches can be performed together with filtering
rules and scope delimitation options. Each run can produce data and
a report that can be easily sent by e-mail or saved into a file system
directory.
The program can parse logs written in RFC 3164 and RFC 5424 formats.
Lograptor requires Python >= 2.7, and is provided with a base configuration for a set
of well known applications. You can easily extend this set adding new applications
with specific pattern search rules.
"""



# -*- coding: utf-8 -*-
"""
Module to manage lograptor applications
"""
#
# Copyright (C), 2011-2017, by SISSA - International School for Advanced Studies.
#
# This file is part of lograptor.
#
# Lograptor is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# file 'LICENSE' in the root directory of the present distribution
# for more details.
#
# @Author Davide Brunato <brunato@sissa.it>
#
import logging
import re
import string
from sre_constants import error as RegexpCompileError

try:
    import configparser
except ImportError:
    # Fall back for Python 2.x
    import ConfigParser as configparser

try:
    import pwd
except ImportError:
    pwd = None

from .exceptions import LogRaptorConfigError, RuleMissingError, LogRaptorOptionError
from .confparsers import AppConfig
from .report import ReportData
from .utils import field_multisub, exact_sub


logger = logging.getLogger(__package__)


class AppRule(object):
    """
    Class to manage application rules. The rules are used to
    parse the log lines and to store matching results.

    Attributes:
        - name: the rule option name in the app configuration file
        - regexp : re.compile object for rule pattern
        - results : dictionary of rule results
        - filter_keys: the filtering keys (all regex groups connected
            to those keys must be not Non to matching a rule)
        - full_match: determine if a rule match represents a full matching
                      for the line (needed for thread matching mode)
        - used_by_report : True if is used by a report rule
        - key_gids : map from gid to result key tuple index
    """

    def __init__(self, name, pattern, args, filter_keys=None):
        """
        Initialize AppRule.

        :param name: the configuration option name
        :param pattern: the option value that represents the search pattern
        :param filter_keys: the filtering keys dictionary if the rule is a filter
        """
        try:
            if not pattern:
                raise LogRaptorConfigError('empty rule %r' % name)
            self.regexp = re.compile(pattern)
        except RegexpCompileError:
            raise LogRaptorConfigError("illegal regex pattern for app\'rule: %r" % name)

        self.name = name
        self.args = args
        self.filter_keys = filter_keys or []
        self.full_match = filter_keys is not None
        self.used_by_report = False
        self.results = dict()

        key_gids = ['host']
        for gid in self.regexp.groupindex:
            if gid != 'host':
                key_gids.append(gid)
        self.key_gids = tuple(key_gids)

        if not self.key_gids:
            raise LogRaptorConfigError("rule gids set empty!")

        self._last_idx = None

    def __repr__(self):
        return u"<%s '%s' at %#x>" % (self.__class__.__name__, self.name, id(self))

    def add_result(self, values):
        """
        Add a tuple or increment the value of an existing one
        in the rule results dictionary.
        """
        idx = [values['host']]
        for gid in self.key_gids[1:]:
            idx.append(values[gid])
        idx = tuple(idx)

        try:
            self.results[idx] += 1
        except KeyError:
            self.results[idx] = 1
        self._last_idx = idx

    def increase_last(self, k):
        """
        Increase the last result by k.
        """
        idx = self._last_idx
        if idx is not None:
            self.results[idx] += k

    def total_events(self, cond, valfld=None):
        """
        Return total number of events in the rule'result set. A condition
        could be provided to select the events to count. If value field (valfld)
        is passed the function compute the sum taking the product of each value with
        correspondent event counter.
        """
        results = self.results

        if cond == "*" and valfld is None:
            return sum(results.values())
        val = self.key_gids.index(valfld) if valfld is not None else None

        if cond == "*":
            tot = 0
            for key in results:
                tot += results[key] * int(key[val])
            return tot
            
        match = re.search("(\w+)(!=|==)\"([^\"]*)\"", cond)
        condpos = self.key_gids.index(match.group(1))
        invert = (match.group(2) == '!=')
        recond = re.compile(match.group(3))

        tot = 0
        for key in results:
            match = recond.search(key[condpos])
            if (not invert and match is not None) or (invert and match is None):
                if valfld is None:
                    tot += results[key]
                else:
                    tot += results[key] * int(key[val])
        return tot

    def top_events(self, num, valfld, usemax, gid):
        """
        Return a list with the top NUM list of events. Each list element
        contain a value, indicating the number of events, and a list of
        matching gid values (usernames, email addresses, clients).
        Instead of calculating the top sum of occurrences a value field
        should be provided to compute the max of a numeric value field or
        the sum of product of value field with events.
        """
        def classify():
            if value is None:
                return

            for j in range(num):
                if top[j] is None:
                    top[j] = [tot, [value]]
                    break
                elif tot == top[j][0]:
                    top[j][1].append(value)
                    break
                elif tot > top[j][0]:
                    top.insert(j, [tot, [value]])
                    break

        if not self.results:
            return []

        results = self.results
        top = [None] * num
        pos = self.key_gids.index(gid)

        # Compute top(max) if a value fld is provided
        if valfld is not None:
            val = self.key_gids.index(valfld)
            if usemax:
                i = 0 
                for key in sorted(results.keys(), key=lambda x: (int(x[val]), x[pos]),
                                  reverse=True)[:num]:
                    top[i] = [int(key[val]), [key[pos]]]
                    i += 1
                return [res for res in top if res is not None]
                
        value = None
        tot = 0
        for key in sorted(results.keys(), key=lambda x: (x[pos])):
            if value is None or value != key[pos]:
                classify()
                value = key[pos]
                tot = results[key] if valfld is None else results[key] * int(key[val])
                continue
            tot += results[key] if valfld is None else results[key] * int(key[val])
        else:
            classify()

        del top[num:]
        return [res for res in top if res is not None]

    def list_events(self, cond, cols, fields):
        """
        Return the list of events, with a specific order and filtered by a condition.
        An element of the list is a tuple with three component. The first is the main
        attribute (first field). The second the second field/label, usually a string
        that identify the service. The third is a dictionary with a key-tuple composed
        by all other fields and values indicating the number of events associated.
        """

        def insert_row():
            """
            Internal function to flush results for a single tabkey to result list.
            """
            row = list(row_template)
            j = 0
            for n in range(cols):
                if row[n] is None:
                    if j == keylen:
                        row[n] = tabvalues
                    else:
                        row[n] = tabkey[j]
                    j += 1
            reslist.append(row)

        if not self.results:
            return []

        # Set local variables
        results = self.results
        pos = [self.key_gids.index(gid) for gid in fields if gid[0] != '"']
        has_cond = cond != "*"

        # If a condition is passed then compile a pattern matching object
        if has_cond:
            match = re.search("(\w+)(!=|==)\"([^\"]*)\"", cond)
            condpos = self.key_gids.index(match.group(1))
            invert = (match.group(2) == '!=')
            recond = re.compile(match.group(3))
        else:
            recond = condpos = None

        # Define the row template with places for values and fixed strings
        row_template = []
        for i in range(cols):
            if fields[i][0] == '"':
                row_template.append(fields[i].strip('"'))
            else:
                row_template.append(None)

        # Set the processing table and reduced key length
        keylen = len(pos) - (len(fields) - cols) - 1
        tabvalues = dict()
        tabkey = None

        reslist = []
        for key in sorted(results, key=lambda x: x[pos[0]]):
            # Skip results that don't satisfy the condition
            if has_cond:
                try:
                    match = recond.search(key[condpos])
                except TypeError:
                    continue
                if ((match is None) and not invert) or ((match is not None) and invert):
                    continue

            new_tabkey = [key[pos[i]] for i in range(keylen)]
            if tabkey is None:
                tabkey = new_tabkey
            elif tabkey != new_tabkey:
                insert_row()
                tabvalues = dict()
                tabkey = [key[pos[i]] for i in range(keylen)]

            value = tuple([key[k] for k in pos[keylen:]])
            if value in tabvalues:
                tabvalues[value] += results[key]
            else:
                tabvalues[value] = results[key]

        if tabvalues:
            insert_row()
        return reslist


class AppLogParser(object):
    """
    Class to manage application log rules and results
    """
    def __init__(self, name, cfgfile, args, logdir, fields, name_cache=None, report=None):
        """
        :param name: application name
        :param cfgfile: application config file
        :param args: cli arguments
        :param logdir: Log directory
        :param fields: Configured fields
        :param name_cache: Optional name cache (--ip-lookup/--uid-lookup/--anonymize options)
        :param report: Optional report (--report option)
        """
        logger.debug('initialize app %r', name)

        self.name = name            # Application name
        self.cfgfile = cfgfile      # App configuration file
        self.args = args
        self.logdir = logdir
        self.fields = fields
        self.name_cache = name_cache

        # Setting instance internal variables for process phase
        self._report = report
        self._thread = args.thread
        self.matches = 0            # Parsed lines counter
        self.unparsed = 0           # Unparsed lines counter
        self._last_rule = None      # Last matched rule
        self._last_idx = None       # Last index matched

        self.config = AppConfig(cfgfiles=cfgfile, appname=name, logdir=logdir)

        self.description = self.config.get('main', 'description')
        self.tags = list(set(re.split('\s*,\s*', self.config.get('main', 'tags'))))
        self._files = list(set(re.split('\s*,\s*', self.config.get('main', 'files'))))
        self.enabled = self.config.getboolean('main', 'enabled')
        self.priority = self.config.getint('main', 'priority')
        self.files = field_multisub(self._files, 'host', args.hosts or ['*'])

        logger.debug('app %r run tags: %r', name, self.tags)
        logger.debug('app %r run files: %r', name, self.files)
        logger.debug('app %r: enabled=%r, priority=%s', name, self.enabled, self.priority)

        self.rules = self.parse_rules()

        if self._report:
            subreports = [sr.name for sr in self._report.subreports]
            self.report_data = [e for e in self.get_report_data() if e.subreport in subreports]

        self.has_filters = any([rule.filter_keys for rule in self.rules])

        if self.has_filters:
            # If the app has filters, reorder rules putting the filters first.
            self.rules = sorted(self.rules, key=lambda x: x.filter_keys)
            logger.debug('filter rules of app %r: %d', name, len(self.filters))
            logger.debug('other rules of app %r: %d', name, len(self.rules) - len(self.filters))
        else:
            for rule in self.rules:
                rule.full_match = True
        logger.info('initialized app %r with %d pattern rules', name, len(self.rules))

    def __repr__(self):
        return u"<%s '%s' at %#x>" % (self.__class__.__name__, self.name, id(self))

    @property
    def filters(self):
        return [rule for rule in self.rules if rule.filter_keys]

    def get_report_data(self):
        report_data = []
        for section in filter(lambda x: x not in ['main', 'rules'], self.config.sections()):
            options = self.config.items(section)
            try:
                data_item = ReportData(section, options, self.rules)
            except RuleMissingError as msg:
                logger.debug(msg)
            except LogRaptorOptionError as err:
                logger.error('skip report data %r for app %r: %s', section, self.name, err)
            else:
                report_data.append(data_item)
        return report_data

    def parse_rules(self):
        """
        Add a set of rules to the app, dividing between filter and other rule set
        """
        # Load patterns: an app is removed when has no defined patterns.
        try:
            rule_options = self.config.items('rules')
        except configparser.NoSectionError:
            raise LogRaptorConfigError("the app %r has no defined rules!" % self.name)

        rules = []
        for option, value in rule_options:
            pattern = value.replace('\n', '')  # Strip newlines for multi-line declarations
            if not self.args.filters:
                # No filters case: substitute the filter fields with the corresponding patterns.
                pattern = string.Template(pattern).safe_substitute(self.fields)
                rules.append(AppRule(option, pattern, self.args))
                continue

            for filter_group in self.args.filters:
                _pattern, filter_keys = exact_sub(pattern, filter_group)
                _pattern = string.Template(_pattern).safe_substitute(self.fields)
                if len(filter_keys) >= len(filter_group):
                    rules.append(AppRule(option, _pattern, self.args, filter_keys))
                elif self._thread:
                    rules.append(AppRule(option, _pattern, self.args))
        return rules

    def increase_last(self, k):
        """
        Increase the counter of the last matched rule by k.
        """
        rule = self._last_rule
        if rule is not None:
            rule.increase_last(k)

    def match_rules(self, log_data):
        """
        Process a log line data message with app's pattern rules.
        Return a tuple with this data:

            Element #0 (app_matched): True if a rule match, False otherwise;
            Element #1 (has_full_match): True if a rule match and is a filter or the
                app has not filters; False if a rule match but is not a filter;
                None otherwise;
            Element #2 (app_thread): Thread value if a rule match and it has a "thread"
                group, None otherwise;
            Element #3 (output_data): Mapping dictionary if a rule match and a map
                of output is requested (--anonymize/--ip/--uid options).
        """
        for rule in self.rules:
            match = rule.regexp.search(log_data.message)
            if match is not None:
                gids = rule.regexp.groupindex
                self._last_rule = rule
                if self.name_cache is not None:
                    values = self.name_cache.match_to_dict(match, rule.key_gids)
                    values['host'] = self.name_cache.map_value(log_data.host, 'host')
                    output_data = {
                        'host': values['host'],
                        'message': self.name_cache.match_to_string(match, gids, values),
                    }
                else:
                    values = {'host': log_data.host}
                    for gid in gids:
                        values[gid] = match.group(gid)
                    output_data = None

                if self._thread and 'thread' in rule.regexp.groupindex:
                    thread = match.group('thread')
                    if rule.filter_keys is not None and \
                            any([values[key] is None for key in rule.filter_keys]):
                        return False, None, None, None
                    if self._report:
                        rule.add_result(values)
                    return True, rule.full_match, thread, output_data
                else:
                    if rule.filter_keys is not None and \
                            any([values[key] is None for key in rule.filter_keys]):
                        return False, None, None, None
                    elif self._report or (rule.filter_keys is not None or not self.has_filters):
                        rule.add_result(values)
                    return True, rule.full_match, None, output_data

        # No rule match: the application log message is not parsable with enabled rules.
        self._last_rule = None
        return False, None, None, None



# -*- coding: utf-8 -*-
"""
This module define classes for building the report produced by a program run.
"""
#
# Copyright (C), 2011-2017, by SISSA - International School for Advanced Studies.
#
# This file is part of lograptor.
#
# Lograptor is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# file 'LICENSE' in the root directory of the present distribution
# for more details.
#
# @Author Davide Brunato <brunato@sissa.it>
#
import logging
import re
import os
import socket
import time
from collections import namedtuple, OrderedDict, MutableMapping
from string import Template

from .info import __version__
from .exceptions import (LogRaptorNoOptionError, LogRaptorNoSectionError, LogRaptorOptionError,
                         RuleMissingError, LogRaptorConfigError)
from . import tui
from .utils import get_fmt_results, htmlsafe, get_value_unit, normalize_path


logger = logging.getLogger(__package__)

TextPart = namedtuple('TextPart', 'fmt text ext')


class ReportData(MutableMapping):
    """
    Class to manage the report items defined for an
    application's logs parsed by lograptor.
    """
    # RE fixed object to check and extract report table parameters
    _color_regexp = re.compile(r'^([a-z]+|#[0-9a-f]{6})', re.IGNORECASE)
    _function_regexp = re.compile(
        r'^(?P<function>table|top|total)(\((\s*(?P<topnum>\d+)\s*,)?'
        r'\s*(?P<headers>(\"[^\"]*\"(\s*,\s*)?)+)\s*\)|)')
    _report_data_regexp = re.compile(
        r'^\(\s*(?P<condition>\*|(?P<field>(\w)+)(!=|==)\"[^\"]*\")\s*,'
        r'\s*(?P<fields>((\w|\"[^\"]*\")(\s*,\s*)?)+)'
        r'(\s*:\s*(?P<add2res>\+)?(?P<valfld>\w+)(\[(?P<unit>(|K|M|G|T)'
        r'(b|bits|B|Bytes))\])?)?\s*\)')
    
    def __init__(self, name, options, rules):
        self._data = OrderedDict()
        self.name = name
        self.subreport = self.title = self.color = self.function = None
        self.rules = dict()
        self.results = []
        self.text = None
        self.html = None
        self.csv = None
        n_headers = 0

        for opt, value in options:
            # Check fixed options
            if opt == 'subreport':
                self.subreport = value
            elif opt == 'title':
                self.title = value
            elif opt == 'color':
                if self._color_regexp.search(value) is None:
                    raise LogRaptorOptionError('color')
                self.color = value
            elif opt == 'function':
                match = self._function_regexp.search(value)
                if not match:
                    raise LogRaptorOptionError('function')

                self.function = match.group('function')
                self.topnum = match.group('topnum')
                self.headers = match.group('headers')
                
                if self.headers:
                    n_headers = len(re.split('\s*,\s*', self.headers))
                else:
                    n_headers = 0
                
                # Parameters value checking
                if self.function == "total" and n_headers > 0:
                    raise LogRaptorOptionError('function', 'function "total" doesn\'t have headers!')
                elif self.function == "top":
                    try:
                        if int(self.topnum) < 0:
                            raise ValueError
                    except (ValueError, TypeError):
                        msg = 'the first argument of function "top" must be a positive integer'
                        raise LogRaptorOptionError('function', msg)
            else:
                # Load and check names of report data options
                for rule in rules:
                    if rule.name == opt or (opt[-1].isdigit() and opt[:-1] == rule.name):
                        rule.used_by_report = True
                        self.rules[opt] = rule
                        break
                else:
                    msg = u"skip report data %r because use undefined or not active app rule!" % self.name
                    raise RuleMissingError(msg)
                self._data[opt] = value

        # Raise if a required option is missing
        if self.subreport is None:
            raise LogRaptorNoOptionError('subreport', name)
        elif self.title is None:
            raise LogRaptorNoOptionError('title', name)
        elif self.function is None:
            raise LogRaptorNoOptionError('function', name)

        # Check the values of report data options
        for opt in self:
            match = self._report_data_regexp.search(self._data[opt])
            if not match:
                raise LogRaptorOptionError(opt, 'a syntax error in report data')

            valfld = match.group('valfld')
            cond = match.group('condition')
            condfield = match.group('field')
            
            fields = re.split('\s*,\s*', match.group('fields'))
            if len(fields) < n_headers:
                raise LogRaptorOptionError('function', 'more headers than fields in the rule!')

            # Check report data function declaration
            if self.function == 'total':
                if cond != '*' and condfield not in self.rules[opt].regexp.groupindex:
                    raise LogRaptorOptionError('function', 'condition %r not in rule %r gids' % (condfield, opt))
                if valfld is not None and valfld not in self.rules[opt].regexp.groupindex:
                    raise LogRaptorOptionError(opt, 'field %r not in rule %r gids' % (valfld, opt))
                if len(fields) > 1:
                    raise LogRaptorOptionError(opt, 'multiple row descriptions!')
                if fields[0][0] != '"':
                    raise LogRaptorOptionError(opt, 'a description must be double-quoted!')
            elif self.function == 'top':
                if valfld is not None and valfld not in self.rules[opt].regexp.groupindex:
                    raise LogRaptorOptionError(opt, 'field %r not in rule %r gids' % (valfld, opt))
                if len(fields) != 1 or fields[0][0] == '"':
                    raise LogRaptorOptionError(opt, 'missing field specification!')
            elif self.function == 'table':
                if valfld is not None:
                    raise LogRaptorOptionError(opt, 'syntax error in report data')
                if cond != '*' and condfield not in self.rules[opt].regexp.groupindex:
                    raise LogRaptorOptionError('function', 'field %r not in rule %r gids' % (condfield, opt))

            # Checking report data fields
            logger.debug('checking fields: %s', fields)
            for field in fields:
                if field[0] == '"' and field[-1] == '"':
                    continue
                if field == "host":
                    continue
                if field not in self.rules[opt].regexp.groupindex:
                    raise LogRaptorOptionError(opt, 'field %r not in rule gids' % field)

    def __repr__(self):
        return u"<%s '%s' at %#x>" % (self.__class__.__name__, self.name, id(self))

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value

    def __delitem__(self, key):
        del self._data[key]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __eq__(self, repitem):
        """
        Compare two 'table' report items. When True the report items
        results are mergeable.
        """
        if self.function != 'table' or repitem.function != 'table':
            return False
        
        if self.title != repitem.title:
            return False
        
        head1 = re.split('\s*,\s*', self.headers)
        head2 = re.split('\s*,\s*', repitem.headers)
        if len(head1) != len(head2):
            return False
        for k in range(len(head1)):
            if head1[k].strip() != head2[k].strip():
                return False
        return True

        # TODO if requested: matching also report item gids
        
    def make_text(self, width):
        """
        Make the text representation of a report data element.
        """
        def mformat(reslist):
            _text = ""
            _buffer = reslist[0]
            for j in range(1, len(reslist)):
                if (_buffer == "") or (len(_buffer) + len(reslist[j])) <= (width - len(filling)):
                    if reslist[j][0] == '[' and reslist[j][-1] == ']':
                        _buffer = '{0} {1}'.format(_buffer, reslist[j])
                    else:
                        _buffer = '{0}, {1}'.format(_buffer, reslist[j])
                else:
                    _text = '{0}{1}\n{2}'.format(_text, _buffer, filling)
                    _buffer = reslist[j]
            _text = '{0}{1}'.format(_text, _buffer)
            return _text

        text = '\n----- {0} -----\n\n'.format(self.title.strip())

        if self.function == 'total':
            width1 = max(len(res[0]) for res in self.results if res is not None)
            for res in self.results:
                padding = ' ' * (width1 - len(res[0]) + 1)
                text = '{0}{1}{2}| {3}\n'.format(text, res[0], padding, res[1])
                
        elif self.function == 'top':
            if self.results[0] is not None:
                width1 = max(len(res[0]) for res in self.results if res is not None)
                width2 = min([width-width1-4,
                              max(len(', '.join(res[1])) for res in self.results if res is not None)])

                text = '{0}{1} | {2}\n'.format(text, ' ' * width1, self.headers.strip('"'))
                text = '{0}{1}-+-{2}-\n'.format(text, '-' * width1, '-' * width2)

                for res in self.results:
                    if res is not None:
                        padding = ' ' * (width1 - len(res[0]) + 1)
                        filling = '{0}| '.format(' ' * (width1 + 1))
                        lastcol = mformat(res[1])
                        text = '{0}{1}{2}| {3}\n'.format(text, res[0], padding, lastcol)
            else:
                text = '{0} {1}\n'.format(text, 'None')
                    
        elif self.function == 'table':
            headers = re.split('\s*,\s*', self.headers)

            colwidth = []
            for i in range(len(headers)-1):
                colwidth.append(max([len(headers[i]), max(len(res[i]) for res in self.results)]))

            for i in range(len(headers)-1):
                text = '{0}{1}{2}| '\
                            .format(text, headers[i].strip('"'), ' ' * (colwidth[i]-len(headers[i])+2))

            text = '{0}{1}\n'.format(text, headers[-1].strip('"'))
            text = '{0}{1}\n'.format(text, '-' * (width-1))

            filling = ""
            for i in range(len(headers)-1):
                filling = '{0}{1}| '.format(filling, ' ' * colwidth[i])
            
            for res in sorted(self.results, key=lambda x: x[0]):
                for i in range(len(headers)-1):
                    text = '{0}{1}{2}| '.format(text, res[i], ' ' * (colwidth[i]-len(res[i])))
                lastcol = get_fmt_results(res[-1], limit=5)
                text = '{0}{1}\n'.format(text, mformat(lastcol))
        self.text = text

    def make_html(self):
        """
        Make the text representation of a report element as html.
        """
        html = None
        if self.function == 'total':
            html = u'<table border="0" width="100%" rules="cols" cellpadding="2">\n'\
                   '<tr><th colspan="2" align="left"><h3><font color="{1}">'\
                   '{0}</font></h3></th></tr>\n'\
                   .format(htmlsafe(self.title.strip()), self.color)

            for res in self.results:
                html = u'{0}<tr><td valign="top" align="right">{1}</td>'\
                       '<td valign="top" width="90%">{2}</td></tr>'\
                       .format(html, res[0], res[1])

        elif self.function == 'top':
            html = u'<table border="0" width="100%" rules="cols" cellpadding="2">\n'\
                   '<tr><th colspan="2" align="left"><h3><font color="{1}">'\
                   '{0}</font></h3></th></tr>\n'\
                   .format(htmlsafe(self.title.strip()), self.color)

            if self.results[0] is not None:
                for res in self.results:
                    if res is not None:
                        html = u'{0}<tr><td valign="top" align="right">{1}</td>'\
                               '<td valign="top" width="90%">{2}</td></tr>'\
                               .format(html, res[0], ', '.join(res[1]))
            else:
                html = u'{0}<tr><td valign="top" align="left">{1}</td>'\
                       .format(html, "None")
                
        elif self.function == 'table':
            html = u'<h3><font color="{1}">{0}</font></h3>'\
                   '<table width="100%" rules="cols" cellpadding="2">\n'\
                   '<tr bgcolor="#aaaaaa">'\
                   .format(htmlsafe(self.title.strip()), self.color)
            
            headers = re.split('\s*,\s*', self.headers)
            for i in range(len(headers)):
                html = '{0}<th align="center" colspan="1">'\
                       '<font color="black">{1}</font></th>'\
                       .format(html, headers[i].strip('"'))

            html = u'{0}</tr>\n'.format(html)
            
            oddflag = False
            lastval = ""            
            for res in sorted(self.results, key=lambda x: x[0]):
                if lastval != res[0]:
                    oddflag = not oddflag
                    if oddflag:
                        html = u'{0}<tr bgcolor="#dddddd">'.format(html)
                    else:
                        html = u'{0}<tr>'.format(html)

                    html = u'{0}<td valign="top" width="15%">{1}</td>'\
                           .format(html, res[0])
                else:
                    if oddflag:
                        html = u'{0}<tr bgcolor="#dddddd">'.format(html)
                    else:
                        html = u'{0}<tr>'.format(html)

                    html = u'{0}<td valign="top" width="15%">&nbsp;</td>'.format(html)
                lastval = res[0]
                
                for i in range(1, len(headers)-1):
                    html = u'{0}<td valign="top" width="15%">{1}</td>'.format(html, res[i])
                lastcol = get_fmt_results(res[-1], limit=10, fmt=u'<font color="darkred">{0}</font>')

                if lastcol[-1].find(u" more skipped]") > -1:
                    html = u'{0}<td valign="top" width="{1}%">{2} {3}</td></tr>\n'\
                           .format(html, 100-15*(len(headers)-1),
                                   u', '.join(lastcol[:-1]), lastcol[-1])
                else:
                    html = u'{0}<td valign="top" width="{1}%">{2}</td></tr>\n'\
                           .format(html, 100-15*(len(headers)-1), u', '.join(lastcol))

        self.html = u'{0}</table>\n<p>\n'.format(html)

    def make_csv(self):
        """
        Get the text representation of a report element as csv.
        """
        import csv
        try:
            from StringIO import StringIO  # Python 2.7
        except ImportError:
            from io import StringIO

        out = StringIO()
        writer = csv.writer(out, delimiter='|', lineterminator='\n', quoting=csv.QUOTE_MINIMAL)

        if self.function == 'total':
            writer.writerows(self.results)                

        elif self.function == 'top':
            rows = [['Value', self.headers.strip('"')]]
            if self.results[0] is not None:
                for res in self.results:
                    if res is not None:
                        rows.append(tuple([res[0], ','.join(res[1])]))
                writer.writerows(rows)
                
        elif self.function == 'table':
            rows = [[header.strip('"') for header in re.split('\s*,\s*', self.headers)]]

            for res in sorted(self.results, key=lambda x: x[0]):
                row = list(res[:-1])
                lastcol = get_fmt_results(res[-1], limit=10)
                if lastcol[-1][0] == '[' and lastcol[-1][-1] == ']':
                    row.append(u'{0} {1}'.format(u', '.join(lastcol[:-1]), lastcol[-1]))
                else:
                    row.append(u', '.join(lastcol))
                rows.append(row)
            writer.writerows(rows)

        self.csv = out.getvalue()

    def parse_report_data(self, opt):
        return self._report_data_regexp.search(self._data[opt])


class Subreport(object):
    """
    Class to manage subreports
    """

    def __init__(self, name, title):
        self.name = name
        self.title = title
        self.report_data = []
        self.reptext = ""

    def __len__(self):
        return len(self.report_data)

    def __repr__(self):
        return u"<%s '%s' at %#x>" % (self.__class__.__name__, self.name, id(self))

    def make(self, apps):
        """
        Make subreport items from results.
        """
        for (appname, app) in sorted(apps.items(), key=lambda x: (x[1].priority, x[0])):
            logger.info('Getting report results from %r', appname)

            for report_data in app.report_data:
                if report_data.subreport != self.name:
                    continue

                if report_data.function == 'total':
                    for opt in report_data:
                        match = report_data.parse_report_data(opt)
                        cond = match.group('condition')
                        valfld = match.group('valfld')
                        unit = match.group('unit')
                        itemtitle = match.group('fields').strip('"')

                        total = report_data.rules[opt].total_events(cond, valfld)
                        if total == 0:
                            continue

                        if unit is not None:
                            total, unit = get_value_unit(total, unit, 'T')
                            total = '{0} {1}'.format(total, unit)
                        else:
                            total = str(total)
                        report_data.results.append(tuple([total, itemtitle]))

                elif report_data.function == 'top':
                    k = int(report_data.topnum)
                    for opt in report_data:
                        match = report_data.parse_report_data(opt)

                        valfld = match.group('valfld')
                        field = match.group('fields')
                        usemax = match.group('add2res') is None

                        toplist = report_data.rules[opt].top_events(k, valfld, usemax, field)
                        report_data.results.extend(toplist)
                        
                elif report_data.function == 'table':
                    cols = len(re.split('\s*,\s*', report_data.headers))
                    for opt in report_data:
                        match = report_data.parse_report_data(opt)
                        cond = match.group('condition')
                        fields = re.split('\s*,\s*', match.group('fields'))
                        tablelist = report_data.rules[opt].list_events(cond, cols, fields)
                        report_data.results.extend(tablelist)

                if report_data.results:
                    self.report_data.append(report_data)

        # Sort and rewrite results as strings with units 
        for report_data in self.report_data:
            if report_data.function == 'top':
                # Sort values
                report_data.results = sorted(report_data.results, key=lambda x: x[0], reverse=True)

                # Get the unit if any and convert numeric results to strings
                unit = None
                for opt in report_data:
                    match = report_data.parse_report_data(opt)
                    unit = match.group('unit')
                    if unit is not None:
                        break

                for res in report_data.results:
                    if unit is not None:
                        v, u = get_value_unit(res[0], unit, 'T')
                        res[0] = '{0} {1}'.format(v, u)
                    else:
                        res[0] = str(res[0])

    def make_format(self, fmt, width):
        """
        Make subreport text in a specified format         
        """
        if not self.report_data:
            return
        
        for data_item in self.report_data:
            if data_item.results:
                if fmt is None or fmt == 'text':
                    data_item.make_text(width)
                elif fmt == 'html':
                    data_item.make_html()
                elif fmt == 'csv':
                    data_item.make_csv()

    def compact_tables(self):
        """
        Compact report items of type "table" with same results type. Report items of type "tables" in the
        same subreport is merged into one. The data are ordered by 1st column.
        """
        items_to_del = set()
        for i in range(len(self.report_data)):
            if i in items_to_del:
                continue
            if self.report_data[i].function[0:5] == 'table':
                for j in range(i+1, len(self.report_data)):
                    if self.report_data[j].function[0:5] == 'table':
                        if self.report_data[i] == self.report_data[j]:
                            logger.debug('Merge of 2 identical report tables: {0}'
                                         .format(self.report_data[i].title)) 
                            items_to_del.add(j)
                            self.report_data[i].results.extend(self.report_data[j].results)
        if items_to_del:
            for i in reversed(sorted(items_to_del, key=lambda x: x)):
                self.report_data.pop(i)

                
class Report(object):
    """
    This helper class holds the contents of a report before it is
    sent to selected channels.
    """
    def __init__(self, name, patterns, args, config):
        self.name = name
        self.patterns = patterns
        self.args = args
        self.config = config

        self.stats = dict()
        self.runtime = time.localtime()

        # Read the report options from the config file
        try:
            options = dict(config.items('%s_report' % name))
        except LogRaptorNoSectionError:
            raise LogRaptorConfigError("no configured report for name %r." % name)

        self.title = Template(options['title']).safe_substitute({
            'localhost': socket.gethostname(),
            'localtime': time.strftime('%c', self.runtime)
        })
        base_path = os.path.dirname(self.config.cfgfile)
        self.html_template = normalize_path(options['html_template'], base_path)
        self.text_template = normalize_path(options['text_template'], base_path)

        self.subreports = [
            Subreport(name=opt.partition('_')[0], title=value)
            for opt, value in options.items() if opt.endswith('_subreport')
        ]

    def make(self, apps):
        """
        Create the report from application results
        """
        for subreport in self.subreports:
            logger.debug('Make subreport "{0}"'.format(subreport.name))
            subreport.make(apps)

        for subreport in self.subreports:
            subreport.compact_tables()

    def cleanup(self):
        pass

    def get_report_parts(self, apps, formats):
        """
        Make report item texts in a specified format.
        """
        for fmt in formats:
            width = 100 if fmt is not None else tui.getTerminalSize()[0]
            for sr in self.subreports:
                sr.make_format(fmt, width)

        logger.debug('Build a map for arguments and run\'s statistics ...')
        value_mapping = {
            'title': self.title,
            'patterns': ', '.join([repr(pattern) for pattern in self.args.patterns]) or None,
            'pattern_files': ', '.join(self.args.pattern_files) or None,
            'hosts': ', '.join(self.args.hosts) or None,
            'apps': u', '.join([
                u'%s(%d)' % (app.name, app.matches) for app in apps.values() if app.matches > 0
            ]),
            'version': __version__
        }

        filters = []
        for flt in self.args.filters:
            filters.append(' AND '.join(['%s=%r' % (k, v.pattern) for k, v in flt.items()]))
        if filters:
            value_mapping['filters'] = ' OR '.join(['(%s)' % item for item in filters])
        else:
            value_mapping['filters'] = filters[0] if filters else None

        value_mapping.update(self.stats)

        report = []
        for fmt in formats:
            if fmt == 'text':
                logger.info('appends a text page report')
                report.append(self.make_text_page(value_mapping))
            elif fmt == 'html':
                logger.info('appends a html page report')
                report.append(self.make_html_page(value_mapping))
            elif fmt == 'csv':
                logger.info('extends with a list of csv subreports')
                report.extend(self.make_csv_tables())
        return report

    def is_empty(self):
        """
        A report is empty when it hasn't subreports or when all subreports are empty.
        """
        return not any(self.subreports)

    def set_stats(self, run_stats):
        """
        Set run statistics for the report.
        """
        self.stats = run_stats.copy()
        self.stats['files'] = ', '.join(self.stats['files'])
        self.stats['tot_files'] = len(run_stats['files'])
        self.stats['extra_tags'] = ', '.join(self.stats['extra_tags'])

    def make_html_page(self, valumap):
        """
        Builds the report as html page, using the template page from file.
        """
        logger.info('Making an html report using template %r.', self.html_template)
        fh = open(self.html_template)
        template = fh.read()
        fh.close()

        parts = []
        for sr in self.subreports:
            report_data = [item.html for item in sr.report_data if item.html]
            if report_data:
                parts.append('\n<h2>{1}</h2>\n'.format(sr.title, sr.reptext))
                parts.extend(report_data)
                parts.append('\n<hr/>')
        
        valumap['subreports'] = '\n'.join(parts) # or "\n<<NO SUBREPORT RELATED EVENTS>>\n"
        html_page = Template(template).safe_substitute(valumap)
        return TextPart(fmt='html', text=html_page, ext='html')

    def make_text_page(self, valumap):
        """
        Builds the report as text page, using the template page from file.
        """
        logger.info('Making a text report page using template %r.', self.text_template)
        fh = open(self.text_template)
        template = fh.read()
        fh.close()
        
        parts = []
        for sr in self.subreports:
            report_data = [item.text for item in sr.report_data if item.text]
            if report_data:
                parts.append('\n{1}\n***** {0} *****\n{1}'.format(sr.title, '*' * (len(sr.title)+12)))
                parts.extend(report_data)

        valumap['subreports'] = '\n'.join(parts) # "\n<<NO SUBREPORT RELATED EVENTS>>\n"
        text_page = Template(template).safe_substitute(valumap)
        return TextPart(fmt='text', text=text_page, ext='txt')

    def make_csv_tables(self):
        """
        Builds the report as a list of csv tables with titles.
        """
        logger.info('Generate csv report tables')
        report_parts = []
        for sr in self.subreports:
            for data_item in sr.report_data:
                report_parts.append(TextPart(fmt='csv', text=data_item.csv, ext='csv'))
        return report_parts



# -*- coding: utf-8 -*-
"""
This module classes and methods for parsing log headers.
"""
#
# Copyright (C), 2011-2017, by SISSA - International School for Advanced Studies.
#
# This file is part of lograptor.
#
# Lograptor is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# file 'LICENSE' in the root directory of the present distribution
# for more details.
#
# @Author Davide Brunato <brunato@sissa.it>
#
import re
from collections import namedtuple
from .exceptions import LogRaptorConfigError


class LogParser(object):
    """
    Base class for building parsers for logs.
    """
    def __init__(self, pattern, app=None):
        """
        Compile the pattern and record group fields. Check if pattern
        include mandatory named groups.
        """
        self.parser = re.compile(pattern)
        self.app = app
        self.LogData = namedtuple('LogData', self.parser.groupindex.keys())
        self.fields = tuple(self.parser.groupindex.keys())

        for field in ('month', 'day', 'ltime', 'message'):
            if field not in self.parser.groupindex:
                msg = '%s: missing mandatory named group "%s"' % (self.__class__.__name__, field)
                raise LogRaptorConfigError(msg)

    def match(self, line):
        return self.parser.match(line)

    def get_data(self, match):
        return self.LogData(*map(match.group, self.fields))


class ParserRFC3164(LogParser):
    """
    Parser and extraction methods for BSD Syslog format (RFC 3164).
    """
    PATTERN = (r'^(?:<(?P<pri>[0-9]{1,3})>|)'
               r'(?P<month>[A-Z,a-z]{3}) (?P<day>(?:[1-3]| )[0-9]) '
               r'(?P<ltime>[0-9]{2}:[0-9]{2}:[0-9]{2}) '
               r'(?:last message repeated (?P<repeat>[0-9]{1,3}) times|'
               r'(?P<host>\S{1,255})\s+'
               r'(?P<message>(?P<apptag>[^ \[\(\:]{1,32})(?:[\[\(\:])?.*))')

    def __init__(self, pattern=None):
        LogParser.__init__(self, pattern or self.PATTERN)
        rfc3164_fields = tuple([
            'pri', 'month', 'day', 'ltime', 'repeat', 'host', 'apptag', 'message'
            ])
        extra = set(self.fields) - set(rfc3164_fields)
        if extra:
            msg = u'no RFC 3164 fields in pattern: {0}'.format(extra)
            raise LogRaptorConfigError(msg)
            

class ParserRFC5424(LogParser):
    """
    Parser for IETF-syslog logs (RFC 5424) .
    """
    PATTERN = (r'^(?:<(?P<pri>[0-9]{1,3})>(?P<ver>[0-9]{0,2}) |)'
               r'(?:-|(?P<year>[0-9]{4})-(?P<month>[0-9]{2})-(?P<day>[0-9]{2})T)'
               r'(?P<ltime>[0-9]{2}:[0-9]{2}:[0-9]{2})(?:|\.(?P<secfrac>[0-9]{1,6}))'
               r'(?:Z |(?P<offset>(?:\+|-)[0-9]{2}:[0-9]{2}) )'
               r'(?:-|(?P<host>\S{1,255})) (?:-|(?P<apptag>\S{1,48})) '
               r'(?:-|(?P<procid>\S{1,128})) (?:-|(?P<msgid>\S{1,32})) '
               r'(?P<message>.*)')

    # The RFC5424 no-value
    NILVALUE = '-'

    def __init__(self, pattern=None):
        LogParser.__init__(self, pattern or self.PATTERN)
        rfc5424_fields = tuple([
            'pri', 'ver', 'year', 'month', 'day', 'ltime', 'secfrac', 'offset',
            'host', 'apptag', 'procid', 'msgid', 'message'
            ])
        extra = set(self.fields) - set(rfc5424_fields)
        if extra:
            msg = u'no RFC 5424 fields in pattern: {0}'.format(extra)
            raise LogRaptorConfigError(msg)


class CycleParsers(object):
    """
    Class that define an iterator for a set of parsers. The additional
    method "detect" permits to find the first parser suitable for the
    argument or return None in alternative.
    """
    PARSERS = [ParserRFC3164(), ParserRFC5424()]

    def __init__(self, parsers=None):
        self.parsers = parsers or self.PARSERS
        self.index = 0
        self.num_parsers = len(self.parsers)

    def __iter__(self):
        return self

    def __next__(self):
        self.index = (self.index + 1) % self.num_parsers
        return self.parsers[self.index]

    def next(self):
        return self.__next__()

    def detect(self, line):
        for i in range(self.num_parsers):
            parser = self.__next__()
            match = parser.match(line)
            if match is not None:
                return parser, match
        else:
            return None, None



# -*- coding: utf-8 -*-
"""
This module contain exception classes for lograptor package.
"""
#
# Copyright (C), 2011-2017, by SISSA - International School for Advanced Studies.
#
# This file is part of lograptor.
#
# Lograptor is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# file 'LICENSE' in the root directory of the present distribution
# for more details.
#
# @Author Davide Brunato <brunato@sissa.it>
#
import logging
try:
    import configparser
except ImportError:
    # Fall back for Python 2.x
    import ConfigParser as configparser


logger = logging.getLogger(__name__)


class LogRaptorException(Exception):
    pass


class LogFormatError(LogRaptorException):
    """
    This exception is raised when there are errors with the
    format of syslog file processed.
    """
    def __init__(self, message):
        Exception.__init__(self, message)
        logger.debug('!FormatError: {0}'.format(message))


class LogRaptorConfigError(LogRaptorException):
    """
    This exception is raised when there are errors in a configuration
    file or when there are misconfiguration problems.
    """
    def __init__(self, message):
        Exception.__init__(self, message)
        logger.debug('!ConfigError: {0}'.format(message))


class LogRaptorArgumentError(LogRaptorException):
    def __init__(self, argument=None, message=None):
        if message is None:
            message = 'syntax error for argument {0}'.format(argument)
        else:
            message = 'argument {0}: {1}'.format(argument, message)
        Exception.__init__(self, message)


class LogRaptorNoSectionError(LogRaptorException, configparser.NoSectionError):
    def __init__(self, section):
        super(LogRaptorNoSectionError, self).__init__(section)


class LogRaptorNoOptionError(LogRaptorException, configparser.NoOptionError):
    def __init__(self, option, section):
        super(LogRaptorNoOptionError, self).__init__(option, section)


class LogRaptorOptionError(LogRaptorException):
    """
    This exception is raised when there is a wrong option values or
    when there are conflicts between options.
    """
    def __init__(self, option, message=None):
        if message is None:
            message = 'syntax error for option %r' % option
        else:
            message = 'option %r: %s' % (option, message)
        Exception.__init__(self, message)
        logger.debug('!OptionError: {0}'.format(message))


class RuleMissingError(LogRaptorException):
    """
    This exception is raised when a rule definition is missing.
    """
    def __init__(self, message):
        Exception.__init__(self, message)
        logger.debug('!RuleMissingError: {0}'.format(message))


class FileMissingError(LogRaptorException):
    """
    This exception is raised when a file is missing.
    """
    def __init__(self, message):
        Exception.__init__(self, message)
        logger.debug('!FileMissingError: {0}'.format(message))


class FileAccessError(LogRaptorException):
    """
    This exception is raised when lograptor has problem to access to a file.
    """
    def __init__(self, message):
        Exception.__init__(self, message)
        logger.debug('!FileAccessError: {0}'.format(message))



# -*- coding: utf-8 -*-
"""
This module contains classes to handle iteration over log files.
"""
#
# Copyright (C), 2011-2017, by SISSA - International School for Advanced Studies.
#
# This file is part of lograptor.
#
# Lograptor is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# file 'LICENSE' in the root directory of the present distribution
# for more details.
#
# @Author Davide Brunato <brunato@sissa.it>
#
import logging
import fnmatch
import ntpath
import glob
import os
import platform
from datetime import datetime
from collections import MutableMapping, OrderedDict
from .timedate import strftimegen

logger = logging.getLogger(__name__)


class GlobDict(MutableMapping):
    """
    A dictionary that uses glob patterns as keys. Includes two additional methods
    glob and iglob to iterate once over dictionary glob patterns, returning couples
    with a path and a list of values.
    """
    def __init__(self, recursive=False, follow_symlinks=False, include=None,
                 exclude=None, exclude_dir=None, dict_class=dict):
        self._data = dict_class()
        self._pathnames = []
        self.recursive = recursive
        self.follow_symlinks = follow_symlinks
        self.include = include or []
        self.exclude = exclude or []
        self.exclude_dir = exclude_dir or []

    def __getitem__(self, path):
        if not isinstance(path, str):
            raise TypeError("path must be a string")
        return self._data[path]

    def __setitem__(self, path, value):
        if not isinstance(path, str):
            raise TypeError("path must be a string")
        if path not in self._data:
            # Update _globs paths
            items = set()
            for pathname in self._pathnames:
                if fnmatch.fnmatch(path, pathname):
                    break
                if fnmatch.fnmatch(pathname, path):
                    items.add(pathname)
            else:
                self._pathnames.append(path)
            if items:
                self._pathnames = [i for i in self._pathnames if i not in items]
        self._data[path] = value

    def __delitem__(self, path):
        for pathname in self._pathnames:
            if fnmatch.fnmatch(path, pathname):
                break
        del self._data[path]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def is_included(self, path):
        basename = ntpath.basename(path)
        dirname = ntpath.dirname(path)
        if any([fnmatch.fnmatch(dirname, pat) for pat in self.exclude_dir]):
            return False
        elif any([fnmatch.fnmatch(basename, pat) for pat in self.exclude]):
            return False
        elif any([fnmatch.fnmatch(basename, pat) for pat in self.include]):
            return True
        else:
            return not self.include

    def iglob(self, path):
        for filename in glob.iglob(path):
            if self.is_included(filename):
                values = {app for k, v in self._data.items() if fnmatch.fnmatch(filename, k) for app in v}
                yield filename, list(values)

    def glob(self, path):
        return list(self.iglob(path))

    def iter_paths(self, pathnames=None, mapfunc=None):
        """
        Special iteration on paths. Yields couples of path and items. If a expanded path
        doesn't match with any files a couple with path and `None` is returned.

        :param pathnames: Iterable with a set of pathnames. If is `None` uses the all \
        the stored pathnames.
        :param mapfunc: A mapping function for building the effective path from various \
        wildcards (eg. time spec wildcards).
        :return: Yields 2-tuples.
        """
        pathnames = pathnames or self._pathnames
        if self.recursive and not pathnames:
            pathnames = ['.']
        elif not pathnames:
            yield []

        if mapfunc is not None:
            for mapped_paths in map(mapfunc, pathnames):
                for path in mapped_paths:
                    if self.recursive and (os.path.isdir(path) or os.path.islink(path)):
                        for t in os.walk(path, followlinks=self.follow_symlinks):
                            for filename, values in self.iglob(os.path.join(t[0], '*')):
                                yield filename, values
                    else:
                        empty_glob = True
                        for filename, values in self.iglob(path):
                            yield filename, values
                            empty_glob = False
                        if empty_glob:
                            yield path, None
        else:
            for path in pathnames:
                if self.recursive and (os.path.isdir(path) or os.path.islink(path)):
                    for t in os.walk(path, followlinks=self.follow_symlinks):
                        for filename, values in self.iglob(os.path.join(t[0], '*')):
                            yield filename, values
                else:
                    empty_glob = True
                    for filename, values in self.iglob(path):
                        yield filename, values
                        empty_glob = False
                    if empty_glob:
                        yield path, None


class FileMap(object):
    """
    A class for building collections of files and for iterating over them.
    """
    def __init__(self, time_period=None, recursive=False, follow_symlinks=False, include=None,
                 exclude=None, exclude_dir=None):
        """
        :param time_period: Time period for filtering the iteration over files. \
        When is `(None, None)` no filter is applied to selected files.
        """
        start_dt, end_dt = time_period or (None, None)
        if start_dt is not None and end_dt is not None and start_dt > end_dt:
            ValueError("start datetime must not be after the end datetime")
        self._filemap = GlobDict(recursive=recursive, follow_symlinks=follow_symlinks, include=include,
                                 exclude=exclude, exclude_dir=exclude_dir, dict_class=OrderedDict)
        self.start_dt = start_dt
        self.end_dt = end_dt

    def __iter__(self):
        """
        Iterate into the file map, with filename glob expansion.
        """
        if self.start_dt is None:
            for path, items in self._filemap.iter_paths():
                yield path, items
        else:
            for path, items in self._filemap.iter_paths(mapfunc=strftimegen(self.start_dt, self.end_dt)):
                if items is None:
                    yield path, items
                elif self.check_stat(path):
                    yield path, items

    def __len__(self):
        return len(list(self.__iter__()))

    def check_stat(self, path):
        """
        Checks logfile stat information for excluding files not in datetime period.
        On Linux it's possible to checks only modification time, because file creation info
        are not available, so it's possible to exclude only older files.
        In Unix BSD systems and windows information about file creation date and times are available,
        so is possible to exclude too newer files.
        """
        statinfo = os.stat(path)
        st_mtime = datetime.fromtimestamp(statinfo.st_mtime)
        if platform.system() == 'Linux':
            check = st_mtime >= self.start_dt
        else:
            st_ctime = datetime.fromtimestamp(statinfo.st_ctime)
            check = st_mtime >= self.start_dt and st_ctime <= self.end_dt

        if not check:
            logger.info("file %r not in datetime period!", path)
        return check
             
    def add(self, files, items):
        """
        Add a list of files with a reference to a list of objects.
        """
        if isinstance(files, (str, bytes)):
            files = iter([files])
        for pathname in files:
            try:
                values = self._filemap[pathname]
            except KeyError:
                self._filemap[pathname] = items
            else:
                values.extend(items)



# -*- coding: utf-8 -*-
"""
Command line tools for lograptor package.
"""
#
# Copyright (C), 2011-2017, by SISSA - International School for Advanced Studies.
#
# This file is part of lograptor.
#
# Lograptor is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# file 'LICENSE' in the root directory of the present distribution
# for more details.
#
# @Author Davide Brunato <brunato@sissa.it>
#
from .core import LogRaptor
from .exceptions import *
from .api import lograptor



# -*- coding: utf-8 -*-
"""
This module contains functions and classes to manage output on text-based
user interface (TUI).
"""
#
# Copyright (C), 2011-2017, by SISSA - International School for Advanced Studies.
#
# This file is part of lograptor.
#
# Lograptor is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# file 'LICENSE' in the root directory of the present distribution
# for more details.
#
# @Author Davide Brunato <brunato@sissa.it>
#
import os
import sys


def getTerminalSize():
    """
    getTerminalSize()
     - get width and height of console
     - works on linux,os x,windows,cygwin(windows)
    """
    import platform
   
    current_os = platform.system()
    tuple_xy = None
    if current_os == 'Windows':
        tuple_xy = _getTerminalSize_windows()
        if tuple_xy is None:
            tuple_xy = _getTerminalSize_tput()
            # needed for window's python in cygwin's xterm!
    if current_os == 'Linux' or current_os == 'Darwin' or current_os.startswith('CYGWIN'):
        tuple_xy = _getTerminalSize_linux()
    if tuple_xy is None:
        tuple_xy = (80, 25)  # default value
    return tuple_xy


def _getTerminalSize_windows():
    from ctypes import windll, create_string_buffer

    # stdin handle is -10
    # stdout handle is -11
    # stderr handle is -12
    handle = windll.kernel32.GetStdHandle(-12)

    try:
        csbi = create_string_buffer(22)
        res = windll.kernel32.GetConsoleScreenBufferInfo(handle, csbi)
    except (IOError, OSError):
        return None

    if res:
        import struct
        (bufx, bufy, curx, cury, wattr,
         left, top, right, bottom, maxx, maxy) = struct.unpack("hhhhHhhhhhh", csbi.raw)
        sizex = right - left + 1
        sizey = bottom - top + 1
        return sizex, sizey
    else:
        return None


def _getTerminalSize_tput():
    # get terminal width
    # src: http://stackoverflow.com/questions/263890/how-do-i-find-the-width-height-of-a-terminal-window
    import subprocess
    try:
        proc = subprocess.Popen(["tput", "cols"], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        output = proc.communicate(input=None)
        cols = int(output[0])
        proc = subprocess.Popen(["tput", "lines"], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        output = proc.communicate(input=None)
        rows = int(output[0])
        return cols, rows
    except (IOError, OSError):
        return None


def _getTerminalSize_linux():

    def ioctl_GWINSZ(fd):
        try:
            import fcntl
            import termios
            import struct
            return struct.unpack('hh', fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234'))
        except (IOError, OSError):
            return None

    cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
    if not cr:
        try:
            f = open(os.ctermid())
            cr = ioctl_GWINSZ(f.fileno())
            f.close()
        except (IOError, OSError):
            pass
    if not cr:
        try:
            cr = (os.environ['LINES'], os.environ['COLUMNS'])
        except KeyError:
            return None
    return int(cr[1]), int(cr[0])


class ProgressBar(object):
    """
    Draw a progress toolbar to stdout. The toolbar is initialized calling
    the function with the first argument set to None.
    """

    def __init__(self, output, maxval=0, suffix=""):
        """
        Create and initialize the progress bar
        """
        self.output = output
        self.suffix = suffix

        self.step_perc = max(1, min(int(100000/maxval), 5))
        self.barwidth = int(0.3 * getTerminalSize()[0])
        self.nextperc = 0

        self.maxval = int(maxval)
        if self.maxval <= 0:
            raise ValueError("Maximum value of a progress bar must be positive number.")

        self.output.write('[{0}] {1} {2}'.format(" " * self.barwidth,
                                                 format(0, '1d'), self.suffix))
        self._width = self.barwidth + 4 + len(self.suffix)
        self.output.flush()
        self.initialized = True

    def redraw(self, value, counter=0):
        """
        Redraw the progress bar with a value and a counter.
        """
        if self.initialized is False:
            return
        if value < 0:
            raise ValueError("Size of a progress bar must be a non negative value")

        perc = int(100 * value / self.maxval)
        if (perc >= self.nextperc) or perc >= 100:
            fill = min(self.barwidth, int(self.barwidth * perc / 100))
            counter = str(counter) if counter > 0 else str(value)

            sys.stdout.write("\b" * self._width)  # return to start of line, after '['
            sys.stdout.write('{0}{1}] {2} {3}'
                             .format("#" * fill, " " * (self.barwidth-fill),
                                     counter, self.suffix))
            sys.stdout.flush()
            if perc < 100:
                self._width = self.barwidth + len(self.suffix) + len(counter) + 3
                self.nextperc = perc + self.step_perc
            else:
                sys.stdout.write("\n")
                self.initialized = False



# -*- coding: utf-8 -*-
"""
This module contain core classes and methods for lograptor package.
"""
#
# Copyright (C), 2011-2017, by SISSA - International School for Advanced Studies.
#
# This file is part of lograptor.
#
# Lograptor is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# file 'LICENSE' in the root directory of the present distribution
# for more details.
#
# @Author Davide Brunato <brunato@sissa.it>
#
import os
import time
import datetime
import re
import glob
import logging
import fileinput
import sys
import fnmatch
from collections import Counter

from .exceptions import (
    LogRaptorConfigError, FileMissingError, LogFormatError, LogRaptorOptionError, LogRaptorArgumentError
)
from .confparsers import LogRaptorConfig
from .application import AppLogParser
from .matcher import create_matcher
from .filemap import FileMap
from .cache import LookupCache
from .dispatchers import UnbufferedDispatcher, LineBufferDispatcher, ThreadedDispatcher
from .report import Report
from .channels import TermChannel, MailChannel, FileChannel
from .timedate import get_datetime_interval
from .utils import is_pipe, is_redirected, protected_property, normalize_path, safe_expand

logger = logging.getLogger(__package__)

try:
    STDIN_FILENO = sys.stdin.fileno()
except ValueError:
    STDIN_FILENO = 0

try:
    prompt = raw_input
except NameError:
    prompt = input


STANDARD_ENCODINGS = ['utf_8', 'latin1', 'latin2']


class LogRaptor(object):
    """
    This is the core class of the lograptor package.

    :param args: Namespace of arguments with run options.
    """
    DEFAULT_CONFIG_FILES = (
        'lograptor.conf',
        os.path.expanduser('~/.config/lograptor/lograptor.conf'),
        '/etc/lograptor/lograptor.conf',
        os.path.join(os.path.dirname(__file__), 'config/lograptor.conf'),
    )

    def __init__(self, args):
        try:
            self.config = LogRaptorConfig(cfgfiles=args.cfgfiles or self.DEFAULT_CONFIG_FILES)
        except (IOError, OSError) as err:
            logger.critical('no configuration available in files %r: %r', args.cfgfiles, err)
            raise FileMissingError('abort %r for previous errors' % __package__)
        else:
            self.args = args
            self.set_logger()
            logger.debug("args={}".format(args))

        # Create a lookup cache when required by arguments
        if any([args.anonymize, args.uid_lookup, args.ip_lookup]):
            self.name_cache = LookupCache(args, self.config)
        else:
            self.name_cache = None

        self._encodings = self.encodings
        self._matcher = self.matcher
        self._patterns = self.patterns
        self._time_period = self.time_period
        self._fields = self.fields
        self._hosts = self.hosts

        # Load applications
        self._config_apps = self._read_apps()
        self._config_tags = {}
        self._apps = self.apps
        self._tags = self.apptags
        self._logmap = self.logmap

        # Setup output channels
        self._channels = self.channels
        self._report = self.report

        if args.loglevel == 4:
            logger.debug("End of lograptor setup!!")
            prompt("press <ENTER> to continue ...")

    def _read_apps(self):
        """
        Read the configuration of applications returning a dictionary

        :return: A dictionary with application names as keys and configuration \
        object as values.
        """
        apps = {}
        for cfgfile in glob.iglob(os.path.join(self.confdir, '*.conf')):
            name = os.path.basename(cfgfile)[0:-5]
            try:
                app = AppLogParser(name, cfgfile, self.args, self.logdir,
                                   self.fields, self.name_cache, self.report)
            except (LogRaptorOptionError, LogRaptorConfigError, LogFormatError) as err:
                logger.error('cannot add app %r: %s', name, err)
            else:
                apps[name] = app

        if not apps:
            raise LogRaptorConfigError('no configured application in %r!' % self.confdir)
        return apps

    # Argument properties
    @property
    def filters(self):
        """Log processor filters."""
        return self.args.filters

    @property
    def recursive(self):
        return self.args.recursive or self.args.dereference_recursive

    @property
    def follow_symlinks(self):
        return self.args.dereference_recursive

    @property
    def include(self):
        return self.args.include

    @property
    def exclude(self):
        if self.exclude_from:
            try:
                exclude = [p.rstrip('\n') for p in fileinput.input(self.args.exclude_from)]
            except (IOError, OSError) as err:
                raise LogRaptorArgumentError('exclude-from', err)
            else:
                return exclude.extend(self.args.exclude)
        else:
            return self.args.exclude

    @property
    def exclude_from(self):
        return self.args.exclude_from

    @property
    def exclude_dir(self):
        return self.args.exclude_dir

    def set_logger(self):
        """
        Setup lograptor logger with an handler and a formatter. The logging
        level is defined by a [0..4] range, where an higher value means a
        more verbose logging. The loglevel value is mapped to correspondent
        logging module's value:

        LOG_CRIT=0 (syslog.h value is 2) ==> logging.CRITICAL
        LOG_ERR=1 (syslog.h value is 3) ==> logging.ERROR
        LOG_WARNING=2 (syslog.h value is 4) ==> logging.WARNING
        LOG_INFO=3 (syslog.h value is 6) ==> logging.INFO
        LOG_DEBUG=4 (syslog.h value is 7) ==> logging.DEBUG

        If the stdout is a tty the log is sent to stderr, otherwise is sent
        to the configured logfile.
        """
        # Higher or lesser argument values are also mapped to DEBUG or CRITICAL
        effective_level = max(logging.DEBUG, logging.CRITICAL - self.args.loglevel * 10)
        logger.setLevel(effective_level)

        # Add an handler if missing
        if not logger.handlers:
            if sys.stdout.isatty():
                handler = logging.StreamHandler()
            else:
                try:
                    handler = logging.FileHandler(self.config.get('main', 'logfile'))
                except (IOError, OSError, TypeError, AttributeError):
                    handler = logging.StreamHandler()
            logger.addHandler(handler)

        # Set the formatter of each handler (normally there is only one handler)
        for handler in logger.handlers:
            if effective_level <= logging.DEBUG:
                formatter = logging.Formatter("[%(levelname)s:%(module)s:%(funcName)s: %(lineno)s] %(message)s")
            else:
                formatter = logging.Formatter("%(levelname)s: %(message)s")
            handler.setLevel(effective_level)
            handler.setFormatter(formatter)

    @protected_property
    def report(self):
        logger.debug("configure a %r report ...", self.args.report)
        if self.args.report is False:
            return False
        elif self.args.report is None:
            return Report('default', self._patterns, self.args, self.config)
        else:
            return Report(self.args.report, self._patterns, self.args, self.config)

    @protected_property
    def patterns(self):
        """
        A tuple with re.RegexObject objects created from regex pattern arguments.
        """
        # No explicit argument for patterns ==> consider the first source argument as pattern.
        if not self.args.patterns and not self.args.pattern_files:
            try:
                self.args.patterns.append(self.args.files.pop(0))
            except IndexError:
                raise LogRaptorArgumentError('PATTERN', 'no search pattern')

        # Get the patterns from arguments and files
        patterns = set()
        if self.args.pattern_files:
            patterns.update([p.rstrip('\n') for p in fileinput.input(self.args.pattern_files)])
        patterns.update(self.args.patterns)
        logger.debug("search patterns to be processed: %r", patterns)

        # If one pattern is empty then skip the other patterns
        if '' in patterns:
            logger.info("an empty pattern provided: match all strings!")
            return tuple()

        try:
            flags = re.IGNORECASE if self.args.case else 0 | re.UNICODE
            return tuple([
                re.compile(r'(\b%s\b)' % pat if self.args.word else '(%s)' % pat, flags=flags)
                for pat in patterns if pat
            ])
        except re.error as err:
            raise LogRaptorArgumentError('wrong regex syntax for pattern: %r' % err)

    @protected_property
    def files(self):
        """
        A list of input sources. Each item can be a file path, a glob path or URL.
        """
        # If no files but a recursion option ==> use the current directory
        if not self.args.files and self.recursive:
            return ['.']
        else:
            return self.args.files

    @protected_property
    def fields(self):
        logger.debug("get fields from arguments ...")
        unknown = [k for item in self.filters for k in item if k not in self.config.options('fields')]
        if unknown:
            raise LogRaptorArgumentError('fields', 'undefined fields: %r.' % list(unknown))

        patterns = {k: v for k, v in self.config.items('patterns')}
        return {
            k: safe_expand(v, patterns) for k, v in self.config.items('fields')
        }

    @protected_property
    def matcher(self):
        """
        Matcher engine: ruled, unruled, unparsed.
        """
        if self.args.matcher is None:
            return 'ruled'
        elif self.args.matcher.startswith('-'):
            matcher = self.args.matcher.strip('-').replace('-', '_')
        else:
            matcher = self.args.matcher

        if matcher not in ['ruled', 'unruled', 'unparsed']:
            raise LogRaptorArgumentError('matcher', 'unknown matcher argument %r' % matcher)
        return matcher

    @protected_property
    def hosts(self):
        hosts = [re.compile(fnmatch.translate(host)) for host in set(self.args.hosts or ['*'])]
        logger.debug('hosts to be processed: %r', hosts)
        return hosts

    @property
    def time_range(self):
        """
        Selected time range for log matching. If `None` then match always (equivalent to 0:00-23:59).
        """
        return self.args.time_range

    @protected_property
    def time_period(self):
        """
        Time period that is determined from the arguments --date and --last. It's a 2-tuple with
        (<start datetime>, <end_datetime>) items. An item is `None` if there isn't a limit.
        """
        if self.args.time_period is None:
            if self.args.files or is_pipe(STDIN_FILENO) or is_redirected(STDIN_FILENO):
                time_period = (None, None)
            else:
                diff = 86400  # 24h = 86400 seconds
                time_period = get_datetime_interval(int(time.time()), diff, 3600)
        else:
            time_period = self.args.time_period

        logger.debug('time period to be processed: %r', time_period)
        return time_period

    @property
    def confdir(self):
        confdir = self.config.get('main', 'confdir')
        return normalize_path(confdir, base_path=os.path.dirname(self.config.cfgfile))

    @property
    def logdir(self):
        confdir = self.config.get('main', 'logdir')
        return normalize_path(confdir, base_path=os.path.dirname(self.config.cfgfile))

    @property
    def encodings(self):
        return self.config.get('main', 'encodings').split(',')

    @protected_property
    def apps(self):
        """
        Dictionary with loaded applications.
        """
        logger.debug("initialize applications ...")
        enabled = None
        apps = self.args.apps or self._config_apps.keys()
        unknown = set(apps) - set(self._config_apps.keys())
        if unknown:
            raise LogRaptorArgumentError("--apps", "not found apps %r" % list(unknown))

        if apps or enabled is None:
            return {k: v for k, v in self._config_apps.items() if k in apps}
        else:
            return {k: v for k, v in self._config_apps.items() if k in apps and v.enabled == enabled}

    @protected_property
    def apptags(self):
        """
        Map from log app-name to an application.
        """
        logger.debug("populate tags map ...")
        apps = self._apps.keys()
        unknown = set(apps)
        unknown.difference_update(self._config_apps.keys())
        if unknown:
            raise ValueError("unknown apps: %r" % list(unknown))

        apps = [v for v in self._config_apps.values() if v.name in apps]
        tagmap = {}
        for app in sorted(apps, key=lambda x: (x.priority, x.name)):
            for tag in app.tags:
                if not tag:
                    raise LogRaptorConfigError('found an empty tag for app %r' % app.name)
                try:
                    tagmap[tag].append(app)
                except KeyError:
                    tagmap[tag] = [app]
        return tagmap

    @protected_property
    def logmap(self):
        apps = sorted(self._apps.values(), key=lambda x: x.priority)
        if self.args.files:
            logmap = FileMap(self._time_period, recursive=self.recursive, follow_symlinks=self.follow_symlinks,
                             include=self.include, exclude=self.exclude, exclude_dir=self.exclude_dir)
            logmap.add(self.args.files, apps)
        elif is_pipe(STDIN_FILENO) or is_redirected(STDIN_FILENO):
            # No files and input by a pipe
            logmap = [(sys.stdin, apps)]
        else:
            # Build the LogMap instance adding the list of files from app config files
            logmap = FileMap(self._time_period, recursive=self.recursive, follow_symlinks=self.follow_symlinks,
                             include=self.include, exclude=self.exclude, exclude_dir=self.exclude_dir)
            for app in apps:
                logmap.add(app.files, [app])

        if self.args.with_filename is None:
            iter_logmap = iter(logmap)
            try:
                next(iter_logmap)
                next(iter_logmap)
            except StopIteration:
                pass
            else:
                # the logmap has more than one file --> prefix log with filename
                self.args.with_filename = True
        return logmap

    @protected_property
    def channels(self):
        """Output channels"""
        try:
            return self._channels
        except AttributeError:
            logger.debug("initialize output channels ...")

        channels = self.args.channels
        config_channels = [sec.rpartition('_')[0] for sec in self.config.sections(suffix='_channel')]
        unknown = set(channels) - set(config_channels)
        if unknown:
            raise ValueError("undefined channel %r" % list(unknown))

        output_channels = []
        for channel in set(channels):
            channel_type = self.config.get('%s_channel' % channel, 'type')
            if channel_type == 'tty':
                output_channels.append(TermChannel(channel, self.args, self.config))
            elif channel_type == 'file':
                output_channels.append(FileChannel(channel, self.args, self.config))
            elif channel_type == 'mail':
                output_channels.append(MailChannel(channel, self.args, self.config))
            else:
                raise LogRaptorConfigError('unknown channel type %r' % channel_type)
        return output_channels

    def __repr__(self):
        return u"<%s %r at %#x>" % (self.__class__.__name__, self.config.cfgfile, id(self))

    def __call__(self, dispatcher=None, parsers=None):
        """
        Log processing main routine. Iterate over the log files calling
        the processing internal routine for each file.
        """
        if dispatcher is None:
            dispatcher = self.create_dispatcher()
        matcher_engine = self.create_matcher(dispatcher, parsers=parsers)
        dispatcher.open()

        files = []
        lines = matches = unknown = 0
        extra_tags = Counter()
        first_event = last_event = None
        if self.args.report:
            self.report.cleanup()

        # Iter between log files. The iteration use the log files modified between the
        # initial and the final date, skipping the other files.
        for (source, apps) in self._logmap:
            if apps is not None:
                logger.debug('process %r for apps %r', source, apps)
            else:
                if self.args.files:
                    logger.error("%s: No such file or directory", source)
                continue

            try:
                for encoding in self._encodings:
                    try:
                        result = matcher_engine(source, apps, encoding)
                    except UnicodeDecodeError:
                        continue
                    break
                else:
                    logger.error("no valid decoder found for %r." % source)
                    continue

                files.append(str(source))

                lines += result.lines
                matches += result.matches
                unknown += result.unknown
                if result.extra_tags:
                    extra_tags.update(result.extra_tags)
                if result.first_event is not None:
                    if first_event is None or first_event > result.first_event:
                        first_event = result.first_event
                if result.last_event is not None:
                    if last_event is None or last_event < result.last_event:
                        last_event = result.last_event

            except IOError as msg:
                if self.args.loglevel:
                    logger.error(msg)

        if not files and self._time_period[0] is not None:
            raise FileMissingError("no file in time period {}!".format([
                datetime.datetime.strftime(e, '%Y-%m-%dT%H:%M:%S') for e in self._time_period
            ]))
        elif not lines:
            return False

        try:
            first_event = datetime.datetime.fromtimestamp(first_event)
            last_event = datetime.datetime.fromtimestamp(last_event)
        except (TypeError, UnboundLocalError):
            first_event = last_event = None

        run_stats = {
            'files': files,
            'first_event': first_event,
            'last_event': last_event,
            'matches': matches,
            'lines': lines,
            'unknown': unknown,
            'extra_tags': extra_tags,
        }

        if unknown > 0:
            logger.warning('found {} lines with an unknown log format'.format(unknown))
        if extra_tags:
            num_lines = sum(extra_tags.values())
            logger.warning(u'found {} unknown extra app tags: {}'.format(num_lines, dict(extra_tags)))

        # If the final report is requested then purge all unmatched threads and set time stamps.
        # Otherwise send final run summary if messages are not disabled.
        if matches > 0 and self.report:
            self.report.set_stats(run_stats)
            self._report.make(self._apps)
            formats = list(set([fmt for channel in self._channels for fmt in channel.formats]))
            report_parts = self._report.get_report_parts(self._apps, formats)
            dispatcher.send_report(report_parts)
        elif self.args.loglevel and not self.args.quiet:
            dispatcher.send_message(self.get_run_summary(run_stats))
        dispatcher.close()

        return matches > 0

    def create_dispatcher(self):
        """
        Return a dispatcher for configured channels.
        """
        before_context = max(self.args.before_context, self.args.context)
        after_context = max(self.args.after_context, self.args.context)

        if self.args.files_with_match is not None or self.args.count or self.args.only_matching or self.args.quiet:
            # Sending of log lines disabled by arguments
            return UnbufferedDispatcher(self._channels)
        elif before_context == 0 and after_context == 0:
            # Don't need line buffering
            return UnbufferedDispatcher(self._channels)
        elif self.args.thread:
            return ThreadedDispatcher(self._channels, before_context, after_context)
        else:
            return LineBufferDispatcher(self._channels, before_context, after_context)

    def create_matcher(self, dispatcher, parsers=None):
        return create_matcher(
            dispatcher=dispatcher,
            parsers=parsers,
            apptags=self.apptags,
            matcher=self.matcher,
            patterns=self.patterns,
            hosts=self.hosts,
            time_range=self.time_range,
            time_period=self.time_period,
            thread=self.args.thread,
            invert=self.args.invert,
            count=self.args.count,
            files_with_match=self.args.files_with_match,
            max_count=self.args.max_count,
            only_matching=self.args.only_matching,
            quiet=self.args.quiet,
            name_cache=self.name_cache,
        )

    def get_config(self):
        """
        Return a formatted text with main configuration parameters.
        """
        # Create a dummy report object if necessary
        channels = [sect.rsplit('_')[0] for sect in self.config.sections(suffix='_channel')]
        channels.sort()
        disabled_apps = [app for app in self._config_apps.keys() if app not in self._apps]
        return u''.join([
            u"\n--- %s configuration ---" % __package__,
            u"\nConfiguration file: %s" % self.config.cfgfile,
            u"\nConfiguration directory: %s" % self.confdir,
            u"\nConfigured applications: %s" % ', '.join(self._config_apps.keys()),
            u"\nDisabled applications: %s" % ', '.join(disabled_apps) if disabled_apps else '',
            u"\nFilter fields: %s" % ', '.join(self.config.options('fields')),
            u"\nOutput channels: %s" % ', '.join(channels) if channels else u'No channels defined',
            u"\nReports: %s\n" % ', '.join(
                [section[:-7] for section in self.config.sections(suffix='_report')]
            ),
            ''
        ])

    def get_run_summary(self, run_stats):
        """
        Produce a text summary from run statistics.

        :param run_stats: A dictionary containing run stats
        :return: Formatted multiline string
        """
        run_stats = run_stats.copy()
        run_stats['files'] = len(run_stats['files'])
        summary = [
            u'\n--- %s run summary ---' % __package__,
            u'Number of processed files: %(files)d',
            u'Total lines read: %(lines)d',
            u'Total log events matched: %(matches)d',
        ]
        if any([app.matches or app.unparsed for app in self.apps.values()]):
            if self.matcher == 'unruled':
                summary.append("Applications found (application rules not used):")
                for app in filter(lambda x: x.matches, self.apps.values()):
                    summary.append(u'  %s(matches=%d)' % (app.name, app.matches))
            else:
                summary.append("Applications found:")
                for app in filter(lambda x: x.matches or x.unparsed, self.apps.values()):
                    summary.append(u'  %s(matches=%d, unparsed=%s)' % (app.name, app.matches, app.unparsed))
        summary.append('\n')
        return '\n'.join(summary) % run_stats



#!/usr/bin/env python
"""
Execute lograptor module as a script (see PEP-338).
"""
#
# Copyright (C), 2011-2017, by SISSA - International School for Advanced Studies.
#
# This file is part of lograptor.
#
# Lograptor is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# file 'LICENSE' in the root directory of the present distribution
# for more details.
#
# @Author Davide Brunato <brunato@sissa.it>
#

if not __package__:
    # When this module is runned without loading the package then
    # __package__ is None or '' and the relative imports are disabled.
    # In this case import the package and set __package__.
    #
    # $ python lograptor --> __package__ == ''
    # $ python lograptor/__main__.py --> __package__ is None
    #
    # Refer to https://www.python.org/dev/peps/pep-0366/ for details.
    import os
    import sys
    pkg_search_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
    if sys.path[0] != pkg_search_path:
        sys.path.insert(0, pkg_search_path)
    import lograptor
    __package__ = lograptor.__name__

from .api import main
main()



# -*- coding: utf-8 -*-
"""
This module contains class to handle events dispatching for lograptor.
"""
#
# Copyright (C), 2011-2017, by SISSA - International School for Advanced Studies.
#
# This file is part of lograptor.
#
# Lograptor is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# file 'LICENSE' in the root directory of the present distribution
# for more details.
#
# @Author Davide Brunato <brunato@sissa.it>
#
from collections import OrderedDict, deque
from itertools import chain, repeat
import abc

try:
    import pwd
except ImportError:
    pwd = None


# noinspection PyUnusedLocal
def dummy(*args, **kwargs):
    return


def create_dispatcher(functions):

    def dispatch(*args, **kwargs):
        for _func in functions:
            _func(*args, **kwargs)

    if not functions:
        return dummy
    elif len(functions) == 1:
        return functions[0]
    else:
        return dispatch


class AbstractDispatcher(object):

    __metaclass__ = abc.ABCMeta

    def __init__(self, channels):
        self.channels = tuple(channels)

    def __setattr__(self, name, value):
        if name == 'channels':
            if not isinstance(value, tuple):
                value = tuple(value)
            self.open = create_dispatcher([channel.open for channel in value])
            self.close = create_dispatcher([channel.close for channel in value])
            self.send_message = create_dispatcher([channel.send_message for channel in value])
            self.send_selected = create_dispatcher([channel.send_selected for channel in value])
            self.send_context = create_dispatcher([channel.send_context for channel in value])
            send_separator = create_dispatcher([channel.send_separator for channel in value])
            self.send_separator = chain([lambda *args: None], repeat(send_separator))
            self.send_report = create_dispatcher([channel.send_report for channel in value])
        super(AbstractDispatcher, self).__setattr__(name, value)

    def dispatch(self, method, *args, **kwargs):
        for channel in self.channels:
            getattr(channel, method)(*args, **kwargs)

    @abc.abstractmethod
    def dispatch_selected(self, *args, **kwargs):
        return

    @abc.abstractmethod
    def dispatch_context(self, *args, **kwargs):
        return

    @abc.abstractmethod
    def reset(self, *args, **kwargs):
        return


class UnbufferedDispatcher(AbstractDispatcher):

    def __init__(self, channels):
        super(UnbufferedDispatcher, self).__init__(channels)

    def __setattr__(self, name, value):
        super(UnbufferedDispatcher, self).__setattr__(name, value)
        if name == "channels":
            self.dispatch_selected = self.send_selected

    def dispatch_selected(self, *args, **kwargs):
        raise NotImplementedError

    def dispatch_context(self, *args, **kwargs):
        return

    def reset(self, *args, **kwargs):
        return


class LineBufferDispatcher(deque, AbstractDispatcher):

    def __init__(self, channels, before_context, after_context):
        super(LineBufferDispatcher, self).__init__(maxlen=before_context)
        AbstractDispatcher.__init__(self, channels)
        self.before_context = before_context
        self.after_context = after_context
        self.last_line = 0
        self.context_until = 0

    def dispatch_selected(self, filename, line_number, match=None, **kwargs):
        next_line = line_number - len(self)
        if self.last_line == 0 or (next_line - self.last_line) > 1:
            next(self.send_separator)()
        for n_line in range(line_number - len(self), line_number):
            self.send_context(
                filename=filename,
                line_number=n_line,
                rawlog=self.popleft(),
                match=match
            )
        self.last_line = line_number
        self.context_until = line_number + self.after_context
        self.send_selected(filename=filename, line_number=line_number, match=match, **kwargs)

    def dispatch_context(self, line_number, rawlog, **kwargs):
        if self.after_context and self.context_until >= line_number:
            self.send_context(line_number=line_number, rawlog=rawlog, **kwargs)
            self.last_line = line_number
        elif self.before_context:
            self.append(rawlog)

    def reset(self):
        self.last_line = 0
        self.context_until = 0
        self.clear()


class ThreadedDispatcher(OrderedDict, AbstractDispatcher):
    """
    A cache for multiple threads.
    """
    def __init__(self, channels, before_context, after_context, max_threads=1000):
        super(ThreadedDispatcher, self).__init__()
        AbstractDispatcher.__init__(self, channels)
        if before_context <= 0:
            raise ValueError("before_context must be a positive integer")
        if after_context <= 0:
            raise ValueError("after_context must be a positive integer")
        self.before_context = before_context
        self.after_context = after_context
        self.context = before_context + after_context + 1
        self.max_threads = max_threads

    def flush(self, key):
        line_cache, matched, after_context = self[key]
        if not matched:
            del self[key]
            return

        next(self.send_separator)()
        for entry in line_cache:
            if entry['match']:
                self.send_selected(**entry)
            else:
                self.send_context(**entry)
        del self[key]

    def dispatch_selected(self, key, **kwargs):
        try:
            line_cache, matched, after_context = self[key]
        except KeyError:
            line_cache = deque()
            line_cache.append(kwargs)
            self[key] = (line_cache, True, 0)
        else:
            if line_cache.maxlen is not None:
                line_cache = deque()
            del self[key]
            line_cache.append(kwargs)
            self[key] = (line_cache, True, 0)

    def dispatch_context(self, key, **kwargs):
        try:
            line_cache, matched, after_context = self[key]
        except KeyError:
            line_cache = deque(maxlen=self.before_context)
            line_cache.append(kwargs)
            self[key] = (line_cache, False, 0)
        else:
            if after_context >= self.after_context:
                self.flush(key)
            else:
                del self[key]
                line_cache.append(kwargs)
                self[key] = (line_cache, matched, 0 if not matched else after_context + 1)

    def reset(self):
        self.clear()



# -*- coding: utf-8 -*-
"""
This module defines communication channels for the program.
"""
#
# Copyright (C), 2011-2017, by SISSA - International School for Advanced Studies.
#
# This file is part of lograptor.
#
# Lograptor is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# file 'LICENSE' in the root directory of the present distribution
# for more details.
#
# @Author Davide Brunato <brunato@sissa.it>
#
import os
import re
import time
import shutil
import socket
import logging
import sys
import tempfile
import abc

from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.utils import formatdate, make_msgid

try:
    from io import BytesIO, StringIO
except ImportError:
    # Python 2 fallback
    from io import StringIO
    BytesIO = StringIO

from .exceptions import LogRaptorConfigError
from .info import __version__
from . import __name__ as package_name
from .utils import mail_message, do_chunked_gzip


logger = logging.getLogger(__package__)


class grep_colors:
    """
    Define a structure for grep color codes.
    """
    DEFAULT_SPEC = "ms=01;31:mc=01;31:sl=:cx=:fn=35:ln=32:bn=32:se=36"
    mt = ms = mc = sl = cx = fn = ln = bn = se = ''

    def __init__(self, color_spec=None):
        color_spec = color_spec or self.DEFAULT_SPEC
        color_dict = dict([item.strip().split('=') for item in self.DEFAULT_SPEC.split(':')])
        try:
            spec_dict = dict([item.strip().split('=') for item in color_spec.split(':')])
        except ValueError as err:
            logger.error("wrong GREP_COLORS spec: %r", err)
        else:
            if 'mt' in spec_dict:
                spec_dict['mc'] = spec_dict['ms'] = spec_dict['mt']
            color_dict.update(spec_dict)

        self.clear = '\033[0m'
        self.rv = bool(color_dict.get('rv', ''))
        self.ne = bool(color_dict.get('ne', ''))

        for attr in ('mt', 'ms', 'mc', 'sl', 'cx', 'fn', 'ln', 'bn', 'se'):
            values = color_dict.get(attr, '') or '0'
            setattr(self, attr, ''.join(['\033[%dm' % int(v) for v in values.split(';')]))


GREP_COLORS = grep_colors(os.environ.get('GREP_COLORS'))


class AbstractChannel(object):
    """
    Abstract base class for lograptor's channels.
    """
    __metaclass__ = abc.ABCMeta

    TEMP_DIR = None

    def __init__(self, name, args, config):
        self.name = name
        self.args = args
        self.config = config
        self.formats = re.split('\s*, \s*', config.get('%s_channel' % name, 'formats'))
        logger.debug('Formats = %r', self.formats)

    def __str__(self):
        return u"<%s '%s'>" % (self.__class__.__name__, self.name)

    def __repr__(self):
        return u"<%s '%s' at %#x>" % (self.__class__.__name__, self.name, id(self))

    def set_tempdir(self):
        if self.TEMP_DIR:
            return
        tmpdir = self.config.get('main', 'tmpdir')
        if tmpdir:
            tempfile.tempdir = tmpdir
        try:
            self.TEMP_DIR = tempfile.mkdtemp('.lograptor')
        except OSError:
            raise LogRaptorConfigError('could not create a temp directory in %r!' % tempfile.tempdir)

    @abc.abstractmethod
    def open(self):
        return

    @abc.abstractmethod
    def close(self):
        return

    @abc.abstractmethod
    def send_message(self, message):
        return

    @abc.abstractmethod
    def send_selected(self, **kwargs):
        return

    @abc.abstractmethod
    def send_context(self, **kwargs):
        return

    @abc.abstractmethod
    def send_separator(self):
        return

    @abc.abstractmethod
    def send_report(self, report_parts):
        return


class TermChannel(AbstractChannel):
    """
    Terminal capable Output Channel
    """
    def __init__(self, name, args, config):
        super(TermChannel, self).__init__(name, args, config)
        if name == 'stdout':
            self._channel = sys.stdout
        color = self.color = args.color == 'always' or args.color == 'auto' and self.isatty()
        invert = GREP_COLORS.rv and self.args.invert
        colon_sep = GREP_COLORS.se + ":" if color else ":"
        dash_sep = GREP_COLORS.se + "-" if color else "-"
        if args.group_separator:
            group_sep = '%s\n' % args.group_separator
            self.group_sep = ''.join([GREP_COLORS.se, group_sep, GREP_COLORS.clear]) if color else group_sep
        else:
            self.group_sep = ''
        selected_color = GREP_COLORS.cx if invert else GREP_COLORS.sl
        context_color = GREP_COLORS.sl if invert else GREP_COLORS.cx

        if invert:
            self.fmt_matching_selected = ''.join([GREP_COLORS.mc, '%s', GREP_COLORS.cx])
            self.fmt_matching_context = ''.join([GREP_COLORS.ms, '%s', GREP_COLORS.sl])
        else:
            self.fmt_matching_selected = ''.join([GREP_COLORS.ms, '%s', GREP_COLORS.sl])
            self.fmt_matching_context = ''.join([GREP_COLORS.mc, '%s', GREP_COLORS.cx])

        fmt_dict = {
            'filename': ''.join([GREP_COLORS.fn, '%(filename)s', GREP_COLORS.clear]) if color else '%(filename)s',
            'line_number': ''.join([GREP_COLORS.ln if color else '', '%(line_number)s']),
            'counter': ''.join([selected_color, '%(counter)s\n', GREP_COLORS.clear]) if color else '%(counter)s\n',
            'selected': ''.join([selected_color, '%(rawlog)s', GREP_COLORS.clear]) if color else '%(rawlog)s',
            'context': ''.join([context_color, '%(rawlog)s', GREP_COLORS.clear]) if color else '%(rawlog)s'
        }

        fmt_parts = []
        if self.args.with_filename:
            fmt_parts.append('filename')
        if self.args.files_with_match is not None:
            self.fmt_selected = '%s\n' % fmt_dict['filename']
            self.fmt_context = ''
        elif self.args.count:
            # When -c/--count option
            fmt_parts.append('counter')
            self.fmt_selected = colon_sep.join([fmt_dict[i] for i in fmt_parts])
            self.fmt_context = ''
        else:
            if self.args.line_number:
                fmt_parts.append('line_number')
            self.fmt_selected = colon_sep.join([fmt_dict[i] for i in fmt_parts + ['selected']])
            self.fmt_context = dash_sep.join(
                [fmt_dict[i] for i in fmt_parts + ['context']]
            )

    def isatty(self):
        try:
            return self._channel.isatty()
        except AttributeError:
            return False

    def open(self):
        return

    def close(self):
        return

    def send_message(self, message):
        self._channel.write(message)

    def send_selected(self, match=None, **kwargs):
        if self.color:
            if match:
                try:
                    kwargs['rawlog'] = ''.join([
                        item if not pos % 2 else self.fmt_matching_selected % item
                        for pos, item in enumerate(match.re.split(kwargs['rawlog']))
                    ])
                except AttributeError:
                    pass
        self._channel.write(self.fmt_selected % kwargs)

    def send_context(self, match=None, **kwargs):
        if self.color:
            if match:
                kwargs['rawlog'] = ''.join([
                    item if not pos % 2 else self.fmt_matching_context % item
                    for pos, item in enumerate(match.re.split(kwargs['rawlog']))
                ])
        self._channel.write(self.fmt_context % kwargs)

    def send_separator(self):
        self._channel.write(self.group_sep)

    def send_report(self, report_parts):
        for fmt in self.formats:
            for part in report_parts:
                if part.fmt == fmt:
                    self._channel.write('\n')
                    self._channel.write(part.text)
                    self._channel.write('\n')


class NoTermChannel(AbstractChannel):

    __metaclass__ = abc.ABCMeta

    def __init__(self, name, args, config):
        super(NoTermChannel, self).__init__(name, args, config)
        self.rawlogs = False
        self.rawfh = None
        self.group_sep = "%s\n" % args.group_separator if args.group_separator else ''

        fmt_dict = {
            'filename': '%(filename)s',
            'line_number': '%(line_number)s',
            'counter': '%(counter)s\n',
            'selected': '%(rawlog)s',
            'context': '%(rawlog)s'
        }
        fmt_parts = []
        colon_sep = ":"
        dash_sep = "-"

        if self.args.with_filename:
            fmt_parts.append('filename')
        if self.args.files_with_match is not None:
            self.fmt_selected = '%s\n' % fmt_dict['filename']
            self.fmt_context = ''
        elif self.args.count:
            # When -c/--count option
            fmt_parts.append('counter')
            self.fmt_selected = colon_sep.join([fmt_dict[i] for i in fmt_parts])
            self.fmt_context = ''
        else:
            if self.args.line_number:
                fmt_parts.append('line_number')
            self.fmt_selected = colon_sep.join([fmt_dict[i] for i in fmt_parts + ['selected']])
            self.fmt_context = dash_sep.join(
                [fmt_dict[i] for i in fmt_parts + ['context']]
            )

    def open(self):
        if self.rawlogs:
            self.rawfh = tempfile.NamedTemporaryFile(mode='w+', delete=False)

    def close(self):
        if self.rawfh is not None:
            self.rawfh.close()


class MailChannel(NoTermChannel):
    """
    Channel type to send results with SMTP.
    """
    def __init__(self, name, args, config):
        super(MailChannel, self).__init__(name, args, config)
        section = '%s_channel' % name
        self.email_address = config.get('main', 'email_address')
        self.smtp_server = config.get('main', 'smtp_server')
        self.mailto = list(set(re.split('\s*, \s*', config.get('%s_channel' % name, 'mailto'))))

        # if self.args.report is not None and self.report.need_rawlogs():
        self.rawlogs = config.getboolean(section, 'include_rawlogs', 'mail_channel')
        if self.rawlogs:
            self.set_tempdir()
            self.rawlogs_limit = config.getint('%s_channel' % name, 'rawlogs_limit') * 1024
        else: 
            self.rawlogs_limit = 0

        self.gpg_encrypt = config.getboolean(section, 'gpg_encrypt', 'mail_channel')
        logger.debug('recipients = %r', self.mailto)
        logger.debug('rawlogs = %r', self.rawlogs)
        logger.debug('rawlogs_limit = %r', self.rawlogs_limit)
        logger.debug('gpg_encrypt = %r', self.gpg_encrypt)

        if self.gpg_encrypt:
            self.gpg_keyringdir = config.get(section, 'gpg_keyringdir')
            gpg_recipients = config.get(section, 'gpg_recipients')
            self.gpg_recipients = [keyid.strip() for keyid in gpg_recipients.split(',') if keyid.strip()]
            gpg_signers = config.get(section, 'gpg_signers')
            self.gpg_signers = [keyid.strip() for keyid in gpg_signers.split(',') if keyid.strip()]
            logger.debug('gpg_keyringdir = %r', self.gpg_recipients)
            logger.debug('gpg_recipients = %r', self.gpg_recipients)
            logger.debug('gpg_signers = %r', self.gpg_signers)

    def __repr__(self):
        return u"<%s '%s: (mailto=%r)' at %#x>" % (self.__class__.__name__, self.name, self.mailto, id(self))

    def send_message(self, message):
        logger.info(message)

    def send_selected(self, **kwargs):
        if self.rawfh:
            self.rawfh.write(self.fmt_selected % kwargs)

    def send_context(self, **kwargs):
        if self.rawfh:
            self.rawfh.write(self.fmt_context % kwargs)

    def send_separator(self, **kwargs):
        if self.rawfh:
            self.rawfh.write(self.group_sep)

    def send_report(self, report_parts):
        """
        Publish by sending the report by e-mail
        """
        logger.info('Creating an email message')
        report_parts = sorted(
            filter(lambda x: x.fmt in self.formats, report_parts),
            key=lambda x: self.formats.index(x.fmt)
        )
        fmtname = '{0}-{1}.{2}' if len(report_parts) > 1 else '{0}.{2}'
        root_part = MIMEMultipart('mixed')
        root_part.preamble = 'This is a multi-part message in MIME format.'

        logger.debug('Creating the text/"text_type" parts')
        for i, text_part in enumerate(report_parts):
            attachment_name = fmtname.format(socket.gethostname(), i, text_part.ext)
            attach_part = MIMEText(text_part.text, text_part.ext, 'utf-8')
            attach_part.add_header('Content-Disposition', 'attachment', filename=attachment_name)
            root_part.attach(attach_part)

        if self.rawlogs:
            out = BytesIO()
            do_chunked_gzip(self.rawfh, out, filename=u'raw.log.gz')
            out.seek(0, os.SEEK_END)
            size = out.tell()

            if size > self.rawlogs_limit:
                logger.warning('%d is over the defined max of %r', size, self.rawlogs_limit)
                logger.warning('Not attaching the raw logs')
            else:
                logger.debug('Creating the application/x-gzip part')
                attach_part = MIMEBase('application', 'x-gzip')
                attach_part.set_payload(out.getvalue())

                from email.encoders import encode_base64

                logger.debug('Encoding the gzipped raw logs with base64')
                encode_base64(attach_part)
                attach_part.add_header('Content-Disposition', 'attachment', filename='raw.log.gz')
                root_part.attach(attach_part)

        if self.gpg_encrypt:
            import gpgme
            try:
                if self.gpg_keyringdir and os.path.exists(self.gpg_keyringdir):
                    logger.debug('Setting keyring dir to %r', self.gpg_keyringdir)
                    os.environ['GNUPGHOME'] = self.gpg_keyringdir

                cleartext = BytesIO(root_part.as_string().encode())
                ciphertext = BytesIO()

                ctx = gpgme.Context()
                ctx.armor = True

                if self.gpg_recipients:
                    recipients = [ctx.get_key(recipient) for recipient in self.gpg_recipients]
                else:
                    recipients = []
                    for key in ctx.keylist():
                        for subkey in key.subkeys:
                            if subkey.can_encrypt:
                                logger.debug('Found can_encrypt key = %d', subkey.keyid)
                                recipients.append(key)
                                break

                signers = [ctx.get_key(signer) for signer in self.gpg_signers]
                if signers:
                    logger.info('Encrypting and signing the report')
                    ctx.signers = signers
                    ctx.encrypt_sign(recipients, gpgme.ENCRYPT_ALWAYS_TRUST, cleartext, ciphertext)
                else:
                    logger.info('Encrypting the report')
                    ctx.encrypt(recipients, gpgme.ENCRYPT_ALWAYS_TRUST, cleartext, ciphertext)

                logger.debug('Creating the MIME envelope for PGP')

                gpg_envelope_part = MIMEMultipart('encrypted')
                gpg_envelope_part.set_param('protocol', 'application/pgp-encrypted', header='Content-Type')
                gpg_envelope_part.preamble = 'This is an OpenPGP/MIME encrypted message (RFC 2440 and 3156)'

                gpg_mime_version_part = MIMEBase('application', 'pgp-encrypted')
                gpg_mime_version_part.add_header('Content-Disposition', 'PGP/MIME version identification')
                gpg_mime_version_part.set_payload('Version: 1')

                gpg_payload_part = MIMEBase('application', 'octet-stream', name='encrypted.asc')
                gpg_payload_part.add_header('Content-Disposition', 'OpenPGP encrypted message')
                gpg_payload_part.add_header('Content-Disposition', 'inline', filename='encrypted.asc')
                gpg_payload_part.set_payload(ciphertext.getvalue())

                gpg_envelope_part.attach(gpg_mime_version_part)
                gpg_envelope_part.attach(gpg_payload_part)

                # envelope becomes the new root part
                root_part = gpg_envelope_part

            except ImportError:
                logger.error('Need crypto libraries for gpg_encrypt.')
                logger.error('Install pygpgme for GPG encryption support.')
                logger.error('Not mailing the report out of caution.')
                return

        # Define headers
        root_part['Date'] = formatdate()
        root_part['From'] = self.email_address
        root_part['To'] = ', '.join(self.mailto)
        root_part['Subject'] = '{0} system events: {1}'.format(
                socket.gethostname(), time.strftime('%c', time.localtime())
        )
        root_part['Message-Id'] = make_msgid()
        root_part['X-Mailer'] = u'{0}-{1}'.format(package_name, __version__)
        
        mail_message(self.smtp_server, root_part.as_string(), self.email_address, self.mailto)
        print('Mailed the report to: {0}'.format(','.join(self.mailto)))


class FileChannel(NoTermChannel):
    """
    FileChannel save results of a run into a set of files
    and directories on the hard drive.
    """
    name = 'file'

    def __init__(self, name, args, config):
        super(FileChannel, self).__init__(name, args, config)
        section = '%s_channel' % name
        self.expire = config.getint(section, 'expire_in')
        self.dirmask = config.get(section, 'dirmask')
        self.filemask = config.get(section, 'filemask')
        self.pubdir = config.get(section, 'pubdir')
        maskmsg = 'Invalid mask for {0}: {1}'

        try:
            self.dirname = time.strftime(self.dirmask, time.localtime())
        except ValueError:
            raise LogRaptorConfigError(maskmsg.format('dirmask', self.dirmask))

        try: 
            self.filename = time.strftime(self.filemask, time.localtime())
        except TypeError:
            LogRaptorConfigError(maskmsg.format('filemask', self.filemask))

        self.rawlogs = config.getboolean(section, 'save_rawlogs')
        if self.rawlogs:
            logger.info('Will save raw logs in the reports directory')

        notify = config.get(section, 'notify')
        self.notify = [addy.strip() for addy in notify.split(',') if addy.strip()]
        if self.notify:
            self.pubroot = config.get(section, 'pubroot')
            logger.debug('pubroot = %r', self.pubroot)
            if not self.pubroot:
                raise LogRaptorConfigError('File channel requires a pubroot when notify is set')
        
        logger.debug('path = %r', self.pubdir)
        logger.debug('filename = %r', self.filename)

    def __repr__(self):
        return u"<%s '%s: pubdir=%r, expire_in=%r' at %#x>" % (
            self.__class__.__name__, self.name, self.pubdir, self.expire, id(self)
        )

    def prune_old(self):
        """
        Removes the directories that are older than a certain date.
        """
        path = self.pubdir
        dirmask = self.dirmask
        expire = self.expire
        expire_limit = int(time.time()) - (86400 * expire)

        logger.info('Pruning directories older than %d days', expire)

        if not os.path.isdir(path):
            logger.warning('Dir %r not found -- skipping pruning', path)
            return

        for entry in os.listdir(path):
            logger.debug('Found: %r', entry)
            if os.path.isdir(os.path.join(path, entry)):
                try: 
                    stamp = time.mktime(time.strptime(entry, dirmask))
                except ValueError as e:
                    logger.info('Dir %r did not match dirmask %r: %r', entry, dirmask, e)
                    logger.info('Skipping %r', entry)
                    continue

                if stamp < expire_limit:
                    shutil.rmtree(os.path.join(path, entry))
                    logger.info('File Publisher: Pruned old dir: %r', entry)
                else:
                    logger.info('%r is still active', entry)
            else:
                logger.info('%r is not a directory. Skipping.', entry)

        logger.info('Finished with pruning')

    def send_message(self, message):
        logger.info(message)

    def send_selected(self, **kwargs):
        if self.rawfh:
            self.rawfh.write(self.fmt_selected % kwargs)

    def send_context(self, **kwargs):
        if self.rawfh:
            self.rawfh.write(self.fmt_context % kwargs)

    def send_separator(self, **kwargs):
        if self.rawfh:
            self.rawfh.write(self.group_sep)

    def send_report(self, report_parts):
        """
        Publish the report parts to local files. Each report part is a text
        with a title and specific extension. For html and plaintext sending
        the report part is unique, for csv send also the stats and unparsed
        string are plain text and report items are csv texts.
        """
        logger.info('Checking and creating the report directory')

        report_parts = sorted(
            filter(lambda x: x.fmt in self.formats, report_parts),
            key=lambda x: self.formats.index(x.fmt)
        )
        workdir = os.path.join(self.pubdir, self.dirname)
        if not os.path.isdir(workdir):
            try: 
                os.makedirs(workdir)
            except OSError as e:
                logger.error('Error creating directory "{0}": {0}'.format(workdir, e))
                return

        fmtname = '{0}-{1}-{2}.{3}' if len(report_parts) > 1 else '{0}-{2}.{3}'

        for i, text_part in enumerate(filter(lambda x: x.fmt in self.formats, report_parts)):
            filename = fmtname.format(self.filename, i, socket.gethostname(), text_part.ext)
            repfile = os.path.join(workdir, filename)
            logger.info('Dumping the report part %d into %r', i, repfile)
            fh = open(repfile, 'w')
            fh.write(text_part.text)
            fh.close()
            print('Report part saved in: %r' % repfile)

        if self.notify:
            logger.info('Creating an email message')
            email_address = self.config.get('main', 'email_address')
            smtp_server = self.config.get('main', 'smtp_server')
            publoc = os.path.join(self.pubroot, self.dirname)

            eml = MIMEText('New lograptor report is available at:\r\n{0}'.format(publoc))
            eml['Subject'] = '{0} system events: {1} (report notification)'.format(
                socket.gethostname(), time.strftime('%c', time.localtime())
            )
            eml['Date'] = formatdate()
            eml['From'] = email_address
            eml['To'] = ', '.join(self.notify)
            eml['X-Mailer'] = u'{0}-{1}'.format(package_name, __version__)

            mail_message(smtp_server, eml.as_string(), email_address, self.notify)
            print('Notification mailed to: {0}'.format(','.join(self.notify)))

        if self.rawlogs:
            logfilename = '{0}.log'.format(self.filename)
            logfile = os.path.join(workdir, '{0}.gz'.format(logfilename))

            logger.info('Gzipping logs and writing them to %r', logfilename)
            outfh = open(logfile, 'w+b')
            do_chunked_gzip(self.rawfh, outfh, logfilename)
            outfh.close()
            print('Gzipped logs saved in: {0}'.format(logfile))

        # Purge old reports
        self.prune_old()



# -*- coding: utf-8 -*-
"""
This module contains various utility functions for lograptor.
"""
#
# Copyright (C), 2011-2017, by SISSA - International School for Advanced Studies.
#
# This file is part of lograptor.
#
# Lograptor is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# file 'LICENSE' in the root directory of the present distribution
# for more details.
#
# @Author Davide Brunato <brunato@sissa.it>
#
import sys
import os
import stat
import string
from functools import wraps
from contextlib import contextmanager

try:
    from urllib.request import urlopen
except ImportError:
    from urllib import urlopen

from .tui import ProgressBar

GZIP_CHUNK_SIZE = 8192


def do_chunked_gzip(infh, outfh, filename):
    """
    A memory-friendly way of compressing the data.
    """
    import gzip

    gzfh = gzip.GzipFile('rawlogs', mode='wb', fileobj=outfh)

    if infh.closed:
        infh = open(infh.name, 'r')
    else:
        infh.seek(0)
        
    readsize = 0
    sys.stdout.write('Gzipping {0}: '.format(filename))

    if os.stat(infh.name).st_size:
        infh.seek(0)
        progressbar = ProgressBar(sys.stdout, os.stat(infh.name).st_size, "bytes gzipped")
        while True:
            chunk = infh.read(GZIP_CHUNK_SIZE)
            if not chunk:
                break

            if sys.version_info[0] >= 3:
                # noinspection PyArgumentList
                gzfh.write(bytes(chunk, "utf-8"))
            else:
                gzfh.write(chunk)

            readsize += len(chunk)
            progressbar.redraw(readsize)

    gzfh.close()


def mail_message(smtp_server, message, from_address, rcpt_addresses):
    """
    Send mail using smtp.
    """
    if smtp_server[0] == '/':
        # Sending the message with local sendmail
        p = os.popen(smtp_server, 'w')
        p.write(message)
        p.close()
    else:
        # Sending the message using a smtp server
        import smtplib

        server = smtplib.SMTP(smtp_server)
        server.sendmail(from_address, rcpt_addresses, message)
        server.quit()


def get_value_unit(value, unit, prefix):
    """
    Return a human-readable value with unit specification. Try to
    transform the unit prefix to the one passed as parameter. When
    transform to higher prefix apply nearest integer round. 
    """
    prefixes = ('', 'K', 'M', 'G', 'T')

    if len(unit):
        if unit[:1] in prefixes:
            valprefix = unit[0] 
            unit = unit[1:]
        else:
            valprefix = ''
    else:
        valprefix = ''
    
    while valprefix != prefix:
        uidx = prefixes.index(valprefix)

        if uidx > prefixes.index(prefix):
            value *= 1024
            valprefix = prefixes[uidx-1]
        else:
            if value < 10240:
                return value, '{0}{1}'.format(valprefix, unit)
            value = int(round(value/1024.0))
            valprefix = prefixes[uidx+1]
    return value, '{0}{1}'.format(valprefix, unit)


def htmlsafe(unsafe):
    """
    Escapes all x(ht)ml control characters.
    """
    unsafe = unsafe.replace('&', '&amp;')
    unsafe = unsafe.replace('<', '&lt;')
    unsafe = unsafe.replace('>', '&gt;')
    return unsafe


def get_fmt_results(results, limit=5, sep='::', fmt=None):
    """
    Return a list of formatted strings representation on a result dictionary.
    The elements of the key are divided by a separator string. The result is
    appended after the key between parentheses. Apply a format transformation
    to odd elements of the key if a fmt parameter is passed.
    """
    result_list = []
    for key in sorted(results, key=lambda x: results[x], reverse=True):
        if len(result_list) >= limit and results[key] <= 1:
            break
        if fmt is not None:
            fmtkey = []
            for i in range(len(key)):
                if i % 2 == 1:
                    fmtkey.append(fmt.format(key[i]))
                else:
                    fmtkey.append(key[i])
            result_list.append(u'{0}({1})'.format(sep.join(fmtkey), results[key]))
        else:
            result_list.append(u'{0}({1})'.format(sep.join(key), results[key]))
    else:
        return result_list
    if fmt is not None:
        result_list.append(fmt.format(u'[%d more skipped]' % (len(results) - len(result_list))))
    else:
        result_list.append(u'[%d more skipped]' % (len(results) - len(result_list)))
    return result_list


def field_multisub(strings, field, values):
    result = set()
    for s in strings:
        result.update(
            [string.Template(s).safe_substitute({field: v}) for v in values]
        )
    return list(result)


def exact_sub(s, mapping):
    fields = list()
    for key, value in mapping.items():
        new_s = string.Template(s).safe_substitute({key: value})
        if new_s != s:
            fields.append(key)
            s = new_s
    return s, fields


def safe_expand(template, mapping):
    """
    Safe string template expansion. Raises an error if the provided substitution mapping has circularities.
    """
    for _ in range(len(mapping) + 1):
        _template = template
        template = string.Template(template).safe_substitute(mapping)
        if template == _template:
            return template
    else:
        raise ValueError("circular mapping provided!")


def results_to_string(results):
    return u', '.join([
        u'%s(%s)' % (key, results[key])
        for key in sorted(results, key=lambda x: results[x], reverse=True)
    ])


def is_pipe(fd):
    return stat.S_ISFIFO(os.fstat(fd).st_mode)


def is_redirected(fd):
    return stat.S_ISREG(os.fstat(fd).st_mode)


def protected_property(func):
    """
    Class method decorator that creates a property that returns the protected attribute
    or the value returned by the wrapped method, if the protected attribute is not defined.
    """
    if func.__name__.startswith('_'):
        raise ValueError("%r: Cannot decorate a protected method!" % func)

    @property
    @wraps(func)
    def proxy_wrapper(self):
        try:
            return getattr(self, '_%s' % func.__name__)
        except AttributeError:
            pass
        return func(self)

    return proxy_wrapper


def normalize_path(path, base_path=None):
    path = path.strip()
    if path.startswith('~/'):
        home = os.path.expanduser("~/")
        return os.path.join(os.path.dirname(home), path[2:])
    elif path.startswith('/') or base_path is None:
        return path
    elif path.startswith('./'):
        return os.path.join(base_path, path[2:])
    else:
        return os.path.abspath(os.path.join(base_path, path))


@contextmanager
def closing(resource):
    try:
        yield resource
    finally:
        resource.close()


def open_resource(source):
    """
    Opens a resource in binary reading mode. Wraps the resource with a
    context manager when it doesn't have one.

    :param source: a filepath or an URL.
    """
    try:
        return open(source, mode='rb')
    except (IOError, OSError) as err:
        try:
            resource = urlopen(source)
        except ValueError:
            pass
        else:
            resource.name = resource.url
            if hasattr(resource, '__enter__'):
                return resource
            else:
                return closing(resource)
        raise err


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Command line interface of the lograptor package.
"""
#
# Copyright (C), 2011-2017, by SISSA - International School for Advanced Studies.
#
# This file is part of lograptor.
#
# Lograptor is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# file 'LICENSE' in the root directory of the present distribution
# for more details.
#
# @Author Davide Brunato <brunato@sissa.it>
#
import sys
import argparse
import time
import re
import sre_constants

from .core import LogRaptor
from .info import __version__, __description__
from .exceptions import (
    LogRaptorConfigError, LogRaptorOptionError, LogFormatError, FileMissingError,
    FileAccessError, LogRaptorArgumentError
)
from .timedate import get_datetime_interval, parse_date_period, parse_last_period, TimeRange


class StoreOptionAction(argparse.Action):
    """
    An action that stores the max length option as value, useful when
    a selection between more conflicting options is needed.
    """
    def __init__(self, option_strings, dest, required=False, help=None, metavar=None):
        const = sorted(option_strings, key=lambda x: len(x))[-1]
        super(StoreOptionAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=0,
            const=const,
            default=None,
            type=str,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        value = getattr(namespace, self.dest, None)
        if value != self.default and value != self.const:
            raise argparse.ArgumentError(self, "conflict with option %s" % value)
        setattr(namespace, self.dest, self.const)


def positive_integer(string):
    value = int(string)
    if value <= 0:
        msg = "%r is not a positive integer" % string
        raise argparse.ArgumentTypeError(msg)
    return value


def filter_spec(string):
    _filter = dict()
    for flt in string.split(','):
        try:
            field, pattern = flt.split('=', 1)
            field, pattern = field.lower(), pattern.strip('\'"')
            if not field:
                raise argparse.ArgumentTypeError('filter %r: empty field name!' % flt)
            elif not pattern:
                raise argparse.ArgumentTypeError('filter %r: empty pattern!' % flt)

            try:
                re.compile(pattern)
                _filter[field] = pattern
            except sre_constants.error:
                raise argparse.ArgumentTypeError("wrong regex pattern in filter %r" % flt)
        except ValueError:
            raise argparse.ArgumentTypeError('filter %r: wrong format!' % flt)
    return _filter


def comma_separated_string(string):
    return [x.strip() for x in string.split(',')]


def last_period_spec(string):
    try:
        diff = parse_last_period(string)
    except TypeError:
        raise argparse.ArgumentTypeError('wrong format: %r' % string)
    else:
        return get_datetime_interval(int(time.time()), diff, 3600)


def date_interval_spec(string):
    try:
        return parse_date_period(string)
    except (TypeError, ValueError):
        raise argparse.ArgumentTypeError('%r: wrong format, use [YYYY]MMDD[,[YYYY]MMDD]' % string)


def create_argument_parser():
    """
    Command line options and arguments parsing. This function return
    a list of options and the list of arguments (pattern, filenames).
    """
    parser = argparse.ArgumentParser(prog='lograptor', description=__description__, add_help=False)
    parser.usage = """%(prog)s [options] PATTERN [FILE ...]
    %(prog)s [options] [-e PATTERN | -f FILE] [FILE ...]
    Try '%(prog)s --help' for more information."""

    group = parser.add_argument_group("General Options")
    group.add_argument(
        "--conf", dest="cfgfiles", action='append', default=None, metavar="FILE",
        help="use a specific configuration file"
    )
    group.add_argument(
        "-d", dest="loglevel", default=2, type=int, metavar="[0-4]", choices=range(5),
        help="Logging level (default is 2, use 4 for debug). A level of 0 suppress also "
             "error messages about nonexistent or unreadable files."
    )
    group.add_argument('-V', '--version', action='version', version=__version__)
    group.add_argument('--help', action='help', help="show this help message and exit")

    group = parser.add_argument_group("Scope Selection")
    group.add_argument(
        "-a", "--apps", metavar='APP[,APP...]', type=comma_separated_string,
        default=[], help="process the log lines related to a list of applications"
    )
    group.add_argument(
        "--hosts", metavar="HOSTNAME/IP[,HOSTNAME/IP...]", type=comma_separated_string,
        default=[], help="process the log lines related to an hostname/IP"
    )
    group.add_argument(
        "-F", "--filter", metavar="FIELD=PATTERN[,FIELD=PATTERN...]",
        action="append", dest="filters", type=filter_spec, default=[],
        help="process the log lines that match all the conditions for rule's field values"
    )
    group.add_argument(
        "--time", metavar="HH:MM,HH:MM", type=TimeRange, action="store", dest="time_range",
        help="process the log lines related to a time range"
    )
    group.add_argument(
        "--date", metavar="[YYYY]MMDD[,[YYYY]MMDD]", action="store", dest="time_period",
        type=date_interval_spec, help="restrict the search scope to a date or a date interval"
    )
    group.add_argument(
        "--last", action="store", dest="time_period", type=last_period_spec,
        metavar="[hour|day|week|month|Nh|Nd|Nw|Nm]",
        help="restrict the search scope to a previous time period"
    )

    group = parser.add_argument_group("Matcher Selection")
    group.add_argument(
        "-G", "--ruled", dest='matcher', action=StoreOptionAction,
        help="use patterns and application rules (default)"
    )
    group.add_argument(
        "-X", "--unruled", dest='matcher', action=StoreOptionAction,
        help="use patterns only, skip application rules"
    )
    group.add_argument(
        "-U", "--unparsed", dest='matcher', action=StoreOptionAction,
        help="match the patterns, don't match any application rule"
    )

    group = parser.add_argument_group("Matching Control")
    group.add_argument(
        "-e", "--regexp", metavar="PATTERN", dest="patterns", default=[],
        action="append", help="use PATTERN for matching"
    )
    group.add_argument(
        "-f", "--file", metavar="FILE", dest="pattern_files", default=[],
        action="append", help="obtain patterns from FILE"
    )
    group.add_argument(
        "-i", "--ignore-case", action="store_true", dest="case", default=False,
        help="ignore case distinctions"
    )
    group.add_argument(
        "-v", "--invert-match", action="store_true", dest="invert", default=False,
        help="invert the sense of patterns regexp matching"
    )
    group.add_argument(
        "-w", "--word-regexp", action="store_true", dest="word", default=False,
        help="force PATTERN to match only whole words"
    )

    group = parser.add_argument_group("General Output Control")
    group.add_argument(
        "--output", default=['stdout'], metavar='CHANNEL[,CHANNEL...]', dest='channels',
        type=comma_separated_string, help="send output to channels (default: ['stdout'])"
    )
    group.add_argument(
        "-c", "--count", action="store_true", default=False,
        help="print only a count of matching lines per FILE"
    )
    group.add_argument(
        "--color", default='auto', nargs='?', choices=['auto', 'always', 'never'],
        help="use markers to highlight the matching strings"
    )
    group.add_argument(
        "-L", "--files-without-match", action="store_false", dest="files_with_match", default=None,
        help="print only names of FILEs containing no match"
    )
    group.add_argument(
        "-l", "--files-with-match", action="store_true", dest="files_with_match",
        help="print only names of FILEs containing matches"
    )
    group.add_argument(
        "-m", "--max-count", metavar='NUM', action="store", type=positive_integer, default=0,
        help="stop after NUM matches"
    )
    group.add_argument(
        "-o", "--only-matching", action="store_true", default=False,
        help="show only the part of a line matching PATTERN"
    )
    group.add_argument(
        "-q", "--quiet", action="store_true", default=False, help="suppress all normal output"
    )
    group.add_argument(
        "-s", "--no-messages", action="store_const", const=0, dest='loglevel',
        help="suppress error messages (equivalent to -d 0)"
    )

    group = parser.add_argument_group("Output Data Control")
    group.add_argument(
        "--report", metavar='NAME', nargs='?', default=False,
        help="produce a report at the end of processing"
    )
    group.add_argument(
        "--ip-lookup", action="store_true", default=False,
        help="translate IP addresses to DNS names"
    )
    group.add_argument(
        "--uid-lookup", action="store_true", default=False,
        help="translate UIDs to usernames"
    )
    group.add_argument(
        "--anonymize", action="store_true", default=False,
        help="anonymize defined rule's fields value"
    )

    group = parser.add_argument_group("Output Line Prefix Control")
    group.add_argument(
        "-n", "--line-number", action="store_true", default=False,
        help="print line number with output lines"
    )
    group.add_argument(
        "-H", "--with-filename", action="store_true", dest="with_filename", default=None,
        help="print the file name for each match"
    )
    group.add_argument(
        "-h", "--no-filename", action="store_false", dest="with_filename", default=None,
        help="suppress the file name prefix on output"
    )

    group = parser.add_argument_group("Context Line Control")
    group.add_argument(
        "-T", "--thread", action="store_true", default=False,
        help="the context is the log thread of the application"
    )
    group.add_argument(
        "-B", "--before-context", metavar='NUM', type=positive_integer, default=0,
        help="print NUM lines of leading context"
    )
    group.add_argument(
        "-A", "--after-context", metavar='NUM', type=positive_integer, default=0,
        help="print NUM lines of trailing context"
    )
    group.add_argument(
        "-C", "--context", metavar='NUM', type=positive_integer, default=0,
        help="print NUM lines of output context"
    )
    group.add_argument(
        "--group-separator", metavar='SEP', default='--',
        help="use SEP as a group separator. By default SEP is double hyphen (--)."
    )
    group.add_argument(
        "--no-group-separator", dest="group_separator", action="store_const", const='',
        help="use empty string as a group separator"
    )

    group = parser.add_argument_group("File and Directory Selection")
    group.add_argument(
        "-r", "--recursive", action="store_true", default=False,
        help="read all files under each directory, recursively"
    )
    group.add_argument(
        "-R", "--dereference-recursive", action="store_true", default=False,
        help="likewise, but follow all symlinks"
    )
    group.add_argument(
        "--include", metavar='GLOB', default=[], action="append",
        help="search only files that match GLOB"
    )
    group.add_argument(
        "--exclude", metavar='GLOB', default=[], action="append",
        help="skip files and directories matching GLOB"
    )
    group.add_argument(
        "--exclude-from", metavar='FILE', default=[], action="append",
        help="skip files matching any file pattern from FILE"
    )
    group.add_argument(
        "--exclude-dir", metavar='DIR', default=[], action="append",
        help="exclude directories matching the pattern DIR"
    )

    parser.add_argument(
        'files', metavar='[FILE ...]', nargs='*',
        help='Input files. Each argument can be a file path or a glob pathname. '
             'A "-" stands for standard input. If no arguments are given then processes '
             'all the files included within the scope of the selected applications.'
    )
    return parser


def has_void_args(argv):
    """
    Check if the command line has no arguments or only the --conf optional argument.
    """
    n_args = len(argv)
    return n_args == 1 or n_args == 2 and argv[1].startswith('--conf=') or n_args == 3 and argv[1] == '--conf'


def lograptor(files, patterns=None, matcher='ruled', cfgfiles=None, apps=None, hosts=None,
              filters=None, time_period=None, time_range=None, case=False, invert=False,
              word=False, files_with_match=None, count=False, quiet=False, max_count=0,
              only_matching=False, line_number=False, with_filename=None,
              ip_lookup=False, uid_lookup=False, anonymize=False, thread=False,
              before_context=0, after_context=0, context=0):
    """
    Run lograptor with arguments. Experimental feature to use the log processor into
    generic Python scripts. This part is still under development, do not use.

    :param files: Input files. Each argument can be a file path or a glob pathname.
    :param patterns: Regex patterns, select the log line if at least one pattern matches.
    :param matcher: Matcher engine, can be 'ruled' (default), 'unruled' or 'unparsed'.
    :param cfgfiles: use a specific configuration file.
    :param apps: process the log lines related to a list of applications.
    :param hosts: process the log lines related to a list of hosts.
    :param filters: process the log lines that match all the conditions for rule's field values.
    :param time_range: process the log lines related to a time range.
    :param time_period: restrict the search scope to a date or a date interval.
    :param case: ignore case distinctions, defaults to `False`.
    :param invert: invert the sense of patterns regexp matching.
    :param word: force PATTERN to match only whole words.
    :param files_with_match: get only names of FILEs containing matches, defaults is `False`.
    :param count: get only a count of matching lines per FILE.
    :param quiet: suppress all normal output.
    :param max_count: stop after NUM matches.
    :param only_matching: get only the part of a line matching PATTERN.
    :param line_number: get line number with output lines.
    :param with_filename: get or suppress the file name for each match.
    :param ip_lookup: translate IP addresses to DNS names.
    :param uid_lookup: translate numeric UIDs to usernames.
    :param anonymize: anonymize defined rule's fields value.
    :param thread: get the lines of logs related to each log line selected.
    :param before_context: get NUM lines of leading context for each log line selected.
    :param after_context: get NUM lines of trailing context for each log line selected.
    :param context: get NUM lines of output context for each log line selected.
    :return:
    """
    cli_parser = create_argument_parser()
    args = cli_parser.parse_args()
    args.files = files
    args.matcher = matcher
    args.cfgfiles = cfgfiles
    args.time_period = time_period
    args.time_range = time_range
    args.case = case
    args.invert = invert
    args.word = word
    args.files_with_match = files_with_match
    args.count = count
    args.quiet = quiet
    args.max_count = max_count
    args.only_matching = only_matching
    args.line_number = line_number
    args.with_filename = with_filename
    args.anonymize = anonymize
    args.ip_lookup = ip_lookup
    args.uid_lookup = uid_lookup
    args.thread = thread
    args.context = context
    args.after_context = after_context
    args.before_context = before_context
    args.patterns = [''] if patterns is None else patterns

    if apps is not None:
        args.apps = apps
    if hosts is not None:
        args.hosts = hosts
    if filters is not None:
        args.filters = filters

    _lograptor = LogRaptor(args)
    return _lograptor()


def main():
    if sys.version_info < (2, 7, 0):
        sys.stderr.write("You need python 2.7 or later to run this program\n")
        sys.exit(1)

    cli_parser = create_argument_parser()
    args = cli_parser.parse_args()

    try:
        if has_void_args(sys.argv) and 'stdout' in args.channels:
            # If the command is called with no relevant args (eg. no args
            # or only --conf argument) then prints the configuration and exit.
            args.patterns.append('')
            _lograptor = LogRaptor(args)
            print(_lograptor.get_config())
            sys.exit(0)

        _lograptor = LogRaptor(args)
        retval = _lograptor()
    except (LogRaptorArgumentError, LogRaptorOptionError, LogRaptorConfigError, LogFormatError,
            FileMissingError, FileAccessError) as err:
        if 'stdout' not in args.channels:
            sys.exit(u"ERROR: {0}\nExiting ...".format(err))
        elif str(err):
            cli_parser.error(err)
        else:
            cli_parser.print_usage()
            sys.exit(2)
    except KeyboardInterrupt:
        print("\nCtrl-C pressed, terminate the process ...")
        sys.exit(1)
    else:
        sys.exit(0 if retval else 1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Setup script for lograptor
"""
#
# Copyright (C), 2011-2017, by SISSA - International School for Advanced Studies.
#
# This file is part of lograptor.
#
# Lograptor is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# file 'LICENSE' in the root directory of the present distribution
# for more details.
#
# @Author Davide Brunato <brunato@sissa.it>
#
import glob
import os
import os.path
import platform
import sys

import setuptools.command.sdist
import setuptools.command.bdist_rpm
from setuptools import setup

import lograptor.info


distro_tags = {
    'centos': 'el',
    'redhat': 'el',
    'Ubuntu': 'ubuntu1'
    }

MAN_SOURCE_DIR = 'doc/_build/man/'
PDF_SOURCE_DIR = 'doc/_build/latex/'


def is_virtual_environment():
    if hasattr(sys, 'real_prefix'):
        # py2/py3 virtualenv environment
        return True
    try:
        # py3 pyvenv environment
        return getattr(sys, 'base_prefix') != sys.prefix
    except AttributeError:
        return False


###
# Set data files depending on the command argument
DOC_FILES = ['README.rst', 'LICENSE', 'doc/lograptor.pdf']
MAN5_FILES = ['man/lograptor.conf.5.gz', 'man/lograptor-apps.5.gz']
MAN8_FILES = ['man/lograptor.8.gz', 'man/lograptor-examples.8.gz']

if 'bdist_rpm' in sys.argv:
    package_data = {}
    data_files = [
        ('/etc/lograptor', [
            'lograptor/config/lograptor.conf',
            'lograptor/config/report_template.html',
            'lograptor/config/report_template.txt']),
        ('/etc/lograptor/conf.d', glob.glob('etc/lograptor/conf.d/*.conf')),
        ('/usr/share/doc/lograptor', DOC_FILES),
        ('/usr/share/man/man5', MAN5_FILES),
        ('/usr/share/man/man8', MAN8_FILES),
    ]
else:
    package_data = {
        'lograptor': [
            'config/lograptor.conf',
            'config/report_template.html',
            'config/report_template.txt',
            'config/conf.d/*.conf',
        ],
    }
    if 'bdist_wheel' in sys.argv:
        data_files = [
            ('share/doc/lograptor', DOC_FILES),
            ('share/man/man5', MAN5_FILES),
            ('share/man/man8', MAN8_FILES),
        ]
    elif is_virtual_environment():
        data_files = [
            ('doc/lograptor', DOC_FILES),
            ('man/man5', MAN5_FILES),
            ('man/man8', MAN8_FILES),
        ]
    else:
        data_files = [
            ('/usr/share/doc/lograptor', DOC_FILES),
            ('/usr/share/man/man5', MAN5_FILES),
            ('/usr/share/man/man8', MAN8_FILES),
        ]


class my_sdist(setuptools.command.sdist.sdist):
    """
    Custom version of sdist command, to update master script
    and compressed version of manual pages.
    """
    def run(self):
        print("Copy {0}lograptor.pdf -> doc/lograptor.pdf".format(PDF_SOURCE_DIR))
        os.system("cp -p {0}lograptor.pdf doc/lograptor.pdf".format(PDF_SOURCE_DIR))
        print("Compress {0}lograptor.8 -> man/lograptor.8.gz".format(MAN_SOURCE_DIR))
        os.system("gzip -c {0}lograptor.8 > man/lograptor.8.gz".format(MAN_SOURCE_DIR))
        print("Compress {0}lograptor.conf.5 -> man/lograptor.conf.5.gz".format(MAN_SOURCE_DIR))
        os.system("gzip -c {0}lograptor.conf.5 > man/lograptor.conf.5.gz".format(MAN_SOURCE_DIR))
        print("Compress {0}lograptor-apps.5 -> man/lograptor-apps.5.gz".format(MAN_SOURCE_DIR))
        os.system("gzip -c {0}lograptor-apps.5 > man/lograptor-apps.5.gz".format(MAN_SOURCE_DIR))
        print("Compress {0}lograptor-examples.8 -> man/lograptor-examples.8.gz".format(MAN_SOURCE_DIR))
        os.system("gzip -c {0}lograptor-examples.8 > man/lograptor-examples.8.gz".format(MAN_SOURCE_DIR))
        setuptools.command.sdist.sdist.run(self)


class my_bdist_rpm(setuptools.command.bdist_rpm.bdist_rpm):

    def _make_spec_file(self):
        """
        Customize spec file inserting %config section
        """
        spec_file = setuptools.command.bdist_rpm.bdist_rpm._make_spec_file(self)
        spec_file.append('%config(noreplace) /etc/lograptor/lograptor.conf')
        spec_file.append('%config(noreplace) /etc/lograptor/report_template.*')
        spec_file.append('%config(noreplace) /etc/lograptor/conf.d/*.conf')
        return spec_file

    def initialize_options(self):
        setuptools.command.bdist_rpm.bdist_rpm.initialize_options(self)
        distro = platform.linux_distribution()
        self.distribution_name = "{0} {1}".format(distro[0], distro[1].split('.')[0])

    def run(self):
        setuptools.command.bdist_rpm.bdist_rpm.run(self)

        msg = 'moving {0} -> {1}'
        print('cd dist/')
        os.chdir('dist')
        files = glob.glob('*-[1-9].noarch.rpm') + glob.glob('*-[1-9][0-9].noarch.rpm')

        distro = platform.linux_distribution(full_distribution_name=False)
        distro_name = distro[0].lower()
        if distro_name in ['centos', 'redhat', 'fedora']:
            if distro_name == 'fedora':
                tag = '.fc' + distro[1].split('.')[0]
            else:
                tag = '.el' + distro[1].split('.')[0]
            for filename in files:
                new_name = filename[:-11] + tag + filename[-11:]
                print(msg.format(filename, new_name))
                os.rename(filename, new_name)

        elif distro_name in ['ubuntu', 'debian']:
            for filename in files:
                print('alien -k {0}'.format(filename))
                os.system('/usr/bin/alien -k {0}'.format(filename))
                print('removing {0}'.format(filename))
                os.unlink(filename)
                if distro[0] == 'Ubuntu':
                    filename = filename[:-11].replace('-', '_', 1) + '_all.deb'
                    new_name = filename[:-8] + 'ubuntu1_all.deb'
                    print(msg.format(filename, new_name))
                    os.rename(filename, new_name)


setup(
    name='lograptor',
    version=lograptor.info.__version__,
    author=lograptor.info.__author__,
    author_email=lograptor.info.__email__,
    description=lograptor.info.__description__,
    license=lograptor.info.__license__,
    maintainer=lograptor.info.__maintainer__,
    long_description=lograptor.info.LONG_DESCRIPTION,
    url='https://github.com/brunato/lograptor',
    packages=['lograptor'],
    package_data=package_data,
    data_files=data_files,
    entry_points={
        'console_scripts': [
            'lograptor=lograptor.api:main'
        ]
    },
    requires=['python (>=2.7)'],
    cmdclass={
        "sdist": my_sdist,
        "bdist_rpm": my_bdist_rpm
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Topic :: Internet :: Log Analysis',
        'Topic :: System :: Systems Administration',
        'Topic :: Text Processing :: Filters',
        'Topic :: Utilities'
    ]
)
